class Stats:
    total_actions = 0
    times_arr = []
    total_actions_time = 0
    error_counter = 0
    objects_existed  = 0
    succeeded_queries = 0
    succeeded_commands = 0

    def __init__(self):
        self.total_actions = 0
        self.times_arr = []
        self.total_actions_time = 0
        self.error_counter = 0
        self.objects_existed = 0
        self.succeeded_queries = 0
        self.succeeded_commands = 0



from aperturedb.Subscriptable import Subscriptable
from aperturedb.transformers.transformer import Transformer
import logging


logger = logging.getLogger(__name__)


class CommonProperties(Transformer):
    """
    This applies some common properties to the data.
    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        """
        Args:
            data: Subscriptable object
            adb_data_source: Data source for the data
            adb_timestamp: Timestamp for the data
            adb_main_object: Main object for the data
        """
        super().__init__(data)

        # Statically set some properties, these are not in the data
        self.adb_data_source = kwargs.get("adb_data_source", None)
        self.adb_timestamp = kwargs.get("adb_timestamp", None)
        self.adb_main_object = kwargs.get("adb_main_object", None)

    def getitem(self, subscript):
        x = self.data[subscript]
        try:
            # x is a transaction that has an add_image command and a blob
            for ic in self._add_image_index:
                src_properties = x[0][ic]["AddImage"]["properties"]
                # Set the static properties, if explicitly set
                if self.adb_data_source:
                    src_properties["adb_data_source"] = self.adb_data_source
                if self.adb_timestamp:
                    src_properties["adb_timestamp"] = self.adb_timestamp
                if self.adb_main_object:
                    src_properties["adb_main_object"] = self.adb_main_object
        except Exception as e:
            logger.exception(e.with_traceback(), stack_info=True)

        return x



import typer

from aperturedb.cli import configure, ingest, utilities, transact

app = typer.Typer()

app.add_typer(ingest.app, name="ingest", help="Ingest data into ApertureDB.")
app.add_typer(configure.app, name="config",
              help="Configure ApertureDB client.")
app.add_typer(utilities.app, name="utils", help="Utilities")
app.add_typer(transact.app, name="transact",
              help="Run a transaction against ApertureDB.")


@app.callback()
def check_context(ctx: typer.Context):
    if ctx.invoked_subcommand != "config":
        configure.check_configured(as_global=False) or \
            configure.check_configured(as_global=True, show_error=True)


if __name__ == "__main__":
    app()



from aperturedb import Subscriptable


class QueryGenerator(Subscriptable.Subscriptable):
    """
    The base class to use for Query Generators.
    """

    def getitem(self, subscript):
        raise Exception("To be implemented in subclass")



class Wrapper():
    """
    This is needed because slicing in Subscriptable returns a list.
    The response handler also needs to be accounted for as
    that will be a part of generator.
    """

    def __init__(self, list, response_handler, strict_response_validation, blobs_relative_to_csv):
        self.list = list
        self.response_handler = response_handler
        self.strict_response_validation = strict_response_validation
        self.blobs_relative_to_csv = blobs_relative_to_csv

    def __len__(self):
        return len(self.list)

    def __getitem__(self, i):
        return self.list[i]


class Subscriptable():
    """
    The base class to use for Data/Generators and such collection types.
    """

    def __getitem__(self, subscript):
        if isinstance(subscript, slice):
            start = subscript.start if subscript.start else 0
            start = len(self) + start if start < 0 else start
            stop = subscript.stop if subscript.stop else len(self)
            step = subscript.step if subscript.step else 1
            wrapper = Wrapper(
                [self.getitem(i) for i in range(start, stop, step)],
                self.response_handler if hasattr(
                    self, "response_handler") else None,
                self.strict_response_validation if hasattr(
                    self, "strict_response_validation") else None,
                self.blobs_relative_to_csv if hasattr(
                    self, "blobs_relative_to_csv") else False
            )
            return wrapper

        else:
            if subscript < len(self):
                return self.getitem(subscript)
            else:
                raise StopIteration()

    def getitem(self, subscript):
        raise Exception("To be implemented in subclass")

    def __iter__(self):
        self.ind = 0
        return self

    def __next__(self):
        if self.ind >= len(self):
            raise StopIteration
        else:
            r = self.getitem(self.ind)
            self.ind += 1
            return r



"""
**Image Objects**
"""

from __future__ import annotations
from typing import Any, Dict, Iterable, List, Tuple, Union
import cv2
import math
import numpy as np

import matplotlib.pyplot as plt

from aperturedb import Utils
from aperturedb.Entities import Entities
from aperturedb.Constraints import Constraints
from ipywidgets import widgets
from IPython.display import display, HTML
import base64
from io import BytesIO
from PIL import Image
from pandas import DataFrame
import logging
logger = logging.getLogger(__name__)


def np_arr_img_to_bytes(arr, format: str = 'JPEG') -> bytes:
    """
    **Convert a NumPy array to bytes**

    Args:
        arr: The NumPy array to convert
        format: The format of the image. Defaults to "JPEG".

    Returns:
        bytes: The image as bytes
    """
    return image_to_bytes(Image.fromarray(arr, 'RGB'), format=format)


def image_to_bytes(image: Image, format: str = "JPEG") -> bytes:
    """
    **Convert an image to bytes**

    Args:
        image (PIL.Image): The image to convert
        format (str, optional): The format of the image. Defaults to "JPEG".

    Returns:
        bytes: The image as bytes
    """
    with BytesIO() as bytes:
        image.save(bytes, format=format)
        return bytes.getvalue()


def rotate(points, angle, c_x=0, c_y=0):
    """
    **Rotate a set of points around a center**

    Args:
        points: The points to rotate; iterable of (x, y) pairs
        angle: The angle of counterclockwise rotation in degrees
        c_x: The x coordinate of the center of rotation
        c_y: The y coordinate of the center of rotation

    Returns:
        points: The rotated points as a NumPy array of shape (n,2) and type int
    """
    ANGLE = np.deg2rad(angle)
    return np.array(
        [
            [
                c_x + np.cos(ANGLE) * (px - c_x) - np.sin(ANGLE) * (py - c_y),
                c_y + np.sin(ANGLE) * (px - c_x) + np.cos(ANGLE) * (py - c_y)
            ]
            for px, py in points
        ]
    ).astype(int)


def resolve(points: np.array, image_meta, operations) -> np.array:
    """
    **Resolve the coordinates of a bounding box to the original image size**

    Given an array of (x,y) points with respect to the coordinates of an image, returns the corresponding coordinates with respect to a the resized or rotated image that resulted from a series of operations.

    Args:
        points (NumPy array): The coordinates of the bounding box; shape (...,2)
        image (dict): The image properties
        operations (list): The operations applied to the image

    Returns:
        resolved_points (NumPy array): The resolved coordinates, same shape as points
    """
    resolved = points.copy()
    if image_meta["adb_image_width"] and image_meta["adb_image_height"]:
        image_meta_width = image_meta["adb_image_width"]
        image_meta_height = image_meta["adb_image_height"]
        # print(f"resolve: {resolved}")
        for operation in operations:
            if operation["type"] == "resize":
                x_ratio = operation["width"] / image_meta_width
                y_ratio = operation["height"] / image_meta_height
                np.multiply(resolved, [x_ratio, y_ratio],
                            out=resolved, casting="unsafe")
                image_meta_width = operation["width"]
                image_meta_height = operation["height"]
            if operation["type"] == "rotate":
                angle = operation["angle"]
                resolved = rotate(
                    resolved, angle, c_x=image_meta_width / 2, c_y=image_meta_height / 2)
            # print(f"resolve: {resolved}")
    else:
        logger.warn(
            "Cannot resolve bounding box coordinates, missing image metadata.")
        logger.warn(
            "Reingest the image with image_properties transformer for this functionality.")
    return resolved


class Images(Entities):
    """
    **The python wrapper of images in ApertureDB.**

    This class serves 2 purposes:
    ***This class is a layer on top of the native query.***
    It facilitate interactions with images in the database in a pythonic way.
    Abstracts the complexity of the query language and the communication with the database.

    ***It includes utility methods to visualize image and annotations**
    Inter convert the representation into NumPy matrices and find similar images,
    related bounding boxes, etc.

    Args:
        db: The database connector, perhaps as returned by `Utils.create_connector`
    """
    db_object = "_Image"

    def inspect(self, use_thumbnails=True) -> Union[Tuple[widgets.IntSlider, widgets.Output], DataFrame]:
        df = super(Images, self).inspect()
        if use_thumbnails:
            def image_base64(im):
                with BytesIO() as buffer:
                    im.save(buffer, 'jpeg')
                    return base64.b64encode(buffer.getvalue()).decode()

            def get_formatters(width):
                def image_formatter(im):
                    return f'<img width={width} style="max-width: 400px" src="data:image/jpeg;base64,{image_base64(im)}" >'

                return image_formatter

            sizer = widgets.IntSlider(min=1, max=400, value=100)
            op = widgets.Output()

            with op:
                image_formatter = get_formatters(sizer.value)
                display(HTML("<div style='max-width: 100%; overflow: auto;'>" +
                             df.to_html(
                                 formatters={'thumbnail': image_formatter}, escape=False)
                             + "</div>")
                        )

            def widget_interaction(c):
                image_formatter = get_formatters(c['new'])
                with op:
                    op.clear_output()
                    display(HTML("<div style='max-width: 100%; overflow: auto;'>" +
                                 df.to_html(
                                     formatters={'thumbnail': image_formatter}, escape=False)
                                 + "</div>")
                            )
            sizer.observe(widget_interaction, 'value')
            return sizer, op

        return df

    def getitem(self, idx):
        item = super().getitem(idx)
        if self.get_image == True:
            if 'thumbnail' not in item:
                buffer = self.get_image_by_index(idx)
                if buffer is not None:
                    item['thumbnail'] = Image.fromarray(
                        self.get_np_image_by_index(idx))
        return item

    # DW interface ends

    def __init__(self, db, batch_size=100, response=None, **kwargs):
        super().__init__(db, response)

        self.db_connector = db

        self.images = {}
        self.images_ids = []
        self.image_sizes = []
        self.images_bboxes = {}
        self.images_polygons = {}
        self.overlays = []
        self.color_for_tag = {}

        self.constraints = None
        self.operations = None
        self.format = None
        self.limit = None
        self.query = None

        self.adjacent = {}

        self.batch_size = batch_size
        self.total_cached_images = 0
        self.display_limit = 20

        self.img_id_prop = "_uniqueid"
        self.bbox_label_prop = "_label"
        self.get_image = True

        if response is not None:
            self.images_ids = list(
                map(lambda x: x[self.img_id_prop], response))

        # Blobs can be passed in addition the response.
        # This would mean that the images are already retrieved.
        # This is useful for usage of the class dor it's utility methods.
        if "blobs" in kwargs:
            blobs = kwargs["blobs"]
            for i, id in enumerate(self.images_ids):
                self.images[id] = blobs[i]

        # If the query is passed, we should save the information
        # for things like operations, constraints, etc.
        # This is useful for resolving the coordinates of bounding boxes.
        if "query" in kwargs:
            self.query = kwargs["query"]

    def __retrieve_batch(self, index):
        '''
        **Retrieve the batch that contains the image with the specified index**
        '''

        # Implement the query to retrieve images
        # for that batch

        total_batches = math.ceil(len(self.images_ids) / self.batch_size)
        batch_id = int(math.floor(index / self.batch_size))

        start = batch_id * self.batch_size
        end = min(start + self.batch_size, len(self.images_ids))

        query = []

        for idx in range(start, end):

            find = {
                "FindImage": {
                    "constraints": {
                        self.img_id_prop: ["==", self.images_ids[idx]]
                    },
                    "blobs": True
                }
            }

            if self.operations and len(self.operations.operations_arr) > 0:
                find["FindImage"]["operations"] = self.operations.operations_arr
            query.append(find)

        res, imgs = self.db_connector.query(query)

        if not self.db_connector.last_query_ok():
            print(self.db_connector.get_last_response_str())
            return

        for (idx, i) in zip(range(start, end), range(end - start)):

            if idx >= len(self.images_ids):
                return

            uniqueid = self.images_ids[idx]
            self.images[str(uniqueid)] = imgs[i]

    def __retrieve_polygons(self, index, constraints, tag_key, tag_format):

        if index > len(self.images_ids):
            print("Index error when retrieving polygons")
            return

        uniqueid = self.images_ids[index]

        query = [{
            "FindImage": {
                "_ref": 1,
                "constraints": {
                    self.img_id_prop: ["==", uniqueid]
                },
                "blobs": False,
                "results": {
                    "list": ["adb_image_width", "adb_image_height"]
                },
            }
        }, {
            "FindPolygon": {
                "image_ref": 1,
                "bounds": True,
                "vertices": True,
            }
        }]

        fpq = query[1]["FindPolygon"]
        if constraints and constraints.constraints:
            fpq["constraints"] = constraints.constraints

        keys_to_add = [tag_key]
        for key in keys_to_add:
            if key == "_label":
                fpq["labels"] = True
            elif key == "_uniqueid":
                fpq["uniqueids"] = True
            elif key == "_area":
                fpq["areas"] = True
            else:
                if "results" not in fpq:
                    fpq["results"] = {}
                fpq_res = fpq["results"]
                if "list" not in fpq_res:
                    fpq_res["list"] = []
                fpq_res["list"].append(key)

        try:
            res, _ = self.db_connector.query(query)

            polygons = []
            bounds = []
            tags = []
            meta = []
            polys = res[1]["FindPolygon"]["entities"]
            operations = self.query["operations"] if self.query and "operations" in self.query else [
            ]
            for poly in polys:
                if tag_key and tag_format:
                    tag = tag_format.format(poly[tag_key])
                    tags.append(tag)
                    meta.append(res[0]["FindImage"]["entities"][0])

                bounds.append(poly["_bounds"])
                converted = []
                for vert in poly["_vertices"]:
                    v = resolve(
                        np.array(vert),
                        res[0]["FindImage"]["entities"][0],
                        operations)
                    converted.append(v)
                polygons.append(converted)

            self.images_polygons[str(uniqueid)] = {
                "bounds": bounds,
                "polygons": polygons,
                "tags": tags,
                "meta": meta
            }

        except Exception as e:
            self.images_polygons[str(uniqueid)] = {
                "bounds": [],
                "polygons": [],
                "tags": [],
                "meta": []
            }
            logger.warn(
                f"Cannot retrieve polygons for image {uniqueid}", exc_info=True)

    def __get_box_from_coords(self, coordinates):
        box = np.array([
            (coordinates["x"], coordinates["y"]),
            (coordinates["x"] +
             coordinates["width"], coordinates["y"]),
            (coordinates["x"] + coordinates["width"],
             coordinates["y"] + coordinates["height"]),
            (coordinates["x"], coordinates["y"] + coordinates["height"])]
        )
        return box

    def __retrieve_bounding_boxes(self, index, constraints):
        # We should fetch all bounding boxes incrementally.
        if self.images_bboxes is None:
            self.images_bboxes = {}

        if index > len(self.images_ids):
            print("Index error when retrieving bounding boxes")
            return

        uniqueid = self.images_ids[index]

        query = [{
            "FindImage": {
                "_ref": 1,
                "constraints": {
                    self.img_id_prop: ["==", uniqueid]
                },
                "results": {
                    "list": ["adb_image_width", "adb_image_height"]
                },
                "blobs": False,
            }
        }, {
            "FindBoundingBox": {
                "image_ref": 1,
                "_ref": 2,
                "blobs": False,
                "coordinates": True,
                "labels": True
            }
        }]
        fbbq = query[1]["FindBoundingBox"]
        if constraints and constraints.constraints:
            fbbq["constraints"] = constraints.constraints

        uniqueid_str = str(uniqueid)
        self.images_bboxes[uniqueid_str] = {}
        try:
            res, images = self.db_connector.query(query)
            bboxes = []
            tags = []
            meta = []
            bounds = []
            if "entities" in res[1]["FindBoundingBox"]:
                for bbox in res[1]["FindBoundingBox"]["entities"]:
                    coordinates = bbox["_coordinates"]
                    box = self.__get_box_from_coords(coordinates)
                    operations = self.query["operations"] if self.query and "operations" in self.query else [
                    ]
                    resolved = resolve(
                        box,
                        # image to bb is 1:n relation
                        res[0]["FindImage"]["entities"][0],
                        operations)
                    bboxes.append(resolved)
                    tags.append(bbox[self.bbox_label_prop])
                    meta.append(res[0]["FindImage"]["entities"][0])
                    bounds.append(box)
        except Exception as e:
            logger.warn(
                f"Cannot retrieve bounding boxes for image {uniqueid}", exc_info=True)
        finally:
            self.images_bboxes[uniqueid_str]["bboxes"] = bboxes
            self.images_bboxes[uniqueid_str]["tags"] = tags
            self.images_bboxes[uniqueid_str]["meta"] = meta
            self.images_bboxes[uniqueid_str]["bounds"] = bounds

    def total_results(self) -> int:
        """
        **Returns the total number of images that matched the query**

        Returns:
            int: Count of images that match the query.
        """

        return len(self.images_ids)

    def get_image_by_index(self, index: int):
        """**Get a single image by its index in the array of retrieved ids**

        Args:
            index (int): Position in the image ids retrieved.

        """

        if index >= len(self.images_ids):
            logger.info("Index is incorrect")
            return

        uniqueid = self.images_ids[index]

        # If image is not retrieved, go and retrieve the batch
        if not str(uniqueid) in self.images:
            self.__retrieve_batch(index)

        if self.images[str(uniqueid)] == None:
            print("Image was not retrieved")

        return self.images[str(uniqueid)]

    def get_np_image_by_index(self, index: int):
        """**Retrieves the NumPy representation of image from database**

        Args:
            index (int): Position in the image ids retrieved.

        """

        image = self.get_image_by_index(index)
        # Just decode the image from buffer
        nparr = np.frombuffer(image, dtype=np.uint8)
        image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        return image

    def get_bboxes_by_index(self, index: int):
        """
        **Get related bounding box for the image**

        Args:
            index (int): Position in the image ids retrieved.

        Returns:
            _type_: _description_
        """
        if not self.images_bboxes or not str(self.images_ids[index]) in self.images_bboxes:
            # Fetch when not present in the map.
            self.__retrieve_bounding_boxes(index, None)

        try:
            bboxes = self.images_bboxes[str(self.images_ids[index])]
        except Exception as e:
            print("Cannot retrieve requested bboxes")
            print(e)

        return bboxes

    # A new search will throw away the results of any previous search
    def search(self, constraints=None, operations=None, format=None, limit=None, sort=None):
        """
        **Sets a new query to retrieve images parameters as described below**

        Args:
            constraints (dict, optional): [Constraints](/python_sdk/parameter_wrappers/Constraints). Defaults to None.
            operations (dict, optional): [Operations](/python_sdk/parameter_wrappers/Operations). Defaults to None.
            format (str, optional): Format of the returned images. Without this, the images are returned as they are stored. Defaults to None.
            limit (int, optional): number of values to restrict results to. Defaults to None.
            sort (dict, optional): [Sort](/python_sdk/parameter_wrappers/Sort). Defaults to None.
        """

        self.constraints = constraints
        self.operations = operations
        self.format = format
        self.limit = limit

        self.images = {}
        self.images_ids = []
        self.images_sizes = []
        self.images_bboxes = {}
        self.images_polygons = {}
        self.overlays = []
        self.color_for_tag = {}

        query = {"FindImage": {}}

        if constraints:
            query["FindImage"]["constraints"] = constraints.constraints

        if format:
            query["FindImage"]["as_format"] = format

        query["FindImage"]["results"] = {}

        if limit:
            query["FindImage"]["results"]["limit"] = limit

        if sort:
            query["FindImage"]["results"]["sort"] = sort

        query["FindImage"]["results"]["list"] = []
        query["FindImage"]["results"]["list"].append(self.img_id_prop)

        # Only retrieve images when needed
        query["FindImage"]["blobs"] = False

        response, images = self.db_connector.query([query])

        try:
            entities = response[0]["FindImage"]["entities"]

            for ent in entities:
                self.images_ids.append(ent[self.img_id_prop])
        except:
            print("Error with search: {}".format(response))

        self.response = entities

    def search_by_property(self, prop_key: str, prop_values: list):
        """
        **Constructs a constraints block and does a new search**

        Args:
            prop_key (str): Key on which to search
            prop_values (list): The values that must match for the key.
        """
        const = Constraints()
        const.is_in(prop_key, prop_values)
        img_sort = {
            "key": prop_key,
            "sequence": prop_values,
        }
        self.search(constraints=const, sort=img_sort)

    def add_polygon_overlay(self, polygons, color=None, alpha=0.4):
        if not color:
            color = self.__random_color()
        self.overlays.append({
            "polygons": polygons,
            "color": color,
            "alpha": alpha
        })

    def add_bbox_overlay(self, bbox, color=None):
        if not color:
            color = self.__random_color()
        self.overlays.append({
            "bbox": self.__get_box_from_coords(bbox),
            "color": color,
        })

    def add_text_overlay(self, text, origin, color=None):
        if not color:
            color = self.__random_color()
        self.overlays.append({
            "text": text,
            "color": color,
            "origin": origin,
        })

    def clear_overlays(self):
        self.overlays = []

    def get_similar_images(self, set_name, n_neighbors):

        imgs_return = Images(self.db_connector)

        for uniqueid in self.images_ids:

            query = [{
                "FindImage": {
                    "_ref": 1,
                    "constraints": {
                        self.img_id_prop:  ["==", uniqueid]
                    },
                    "blobs": False,
                }
            }, {
                "FindDescriptor": {
                    "set": set_name,
                    "is_connected_to": {
                        "ref": 1,
                    },
                    "blobs": True
                }
            }]

            response, blobs = self.db_connector.query(query)

            query = [{
                "FindDescriptor": {
                    "_ref": 1,
                    "set": set_name,
                    "k_neighbors": n_neighbors + 1,
                    "blobs":     False,
                    "distances": True,
                    "uniqueids": True,
                }
            }, {
                "FindImage": {
                    "is_connected_to": {
                        "ref": 1,
                    },
                    "group_by_source": True,
                    "results": {
                        "list": [self.img_id_prop]
                    }
                }
            }]

            response, blobs = self.db_connector.query(query, blobs)

            try:
                descriptors = response[0]["FindDescriptor"]["entities"]
                ordered_descs_ids = [x["_uniqueid"] for x in descriptors]

                # Images are not sorted by distance, so we need to sort them
                # That's why we use "group_by_source":
                # To have a mapping between descriptors and associated images.
                imgs_map = response[1]["FindImage"]["entities"]

                for desc_id in ordered_descs_ids:
                    img_info = imgs_map[desc_id][0]
                    # We could assert here that there is only one image per descriptor.
                    imgs_return.images_ids.append(img_info[self.img_id_prop])

            except Exception as e:
                print("Error with search: {}".format(response))
                print(e)
                print("Error with similarity search")

        return imgs_return

    def __draw_text_with_shadow(self, image, text, origin, color, typeface=cv2.FONT_HERSHEY_SIMPLEX, scale=0.75, thickness=2, shadow_color=None, shadow_radius=4, meta=None):
        width, height = image.shape[1], image.shape[0]
        thickness = max(1, int(min(width, height) / 200))
        if not shadow_color:
            shadow_color = self.__contrasting_color(color)

        scale = thickness / 3
        resolved_origin = np.array([origin])
        operations = self.query["operations"] if self.query and "operations" in self.query else [
        ]
        if meta and "adb_image_width" in meta and "adb_image_height" in meta:
            resolved_origin = resolve(resolved_origin, meta, operations)
        else:
            logger.warning(
                f"Cannot resolve text origin, missing image metadata in {meta}.")
        cv2.putText(image, text, resolved_origin[0], typeface, scale,
                    shadow_color, thickness + shadow_radius, cv2.LINE_AA)

        cv2.putText(image, text, resolved_origin[0], typeface,
                    scale, color, thickness, cv2.LINE_AA)

    def __draw_bbox(self, image, bbox, color, thickness=1):
        width, height = image.shape[1], image.shape[0]
        thickness = max(1, int(min(width, height) / 200))
        cv2.drawContours(image, [bbox], 0, color, thickness)

    def __get_color_for_tag(self, tag: str) -> Tuple[int, int, int]:
        if tag not in self.color_for_tag:
            color = self.__random_color()
            self.color_for_tag[tag] = color
        return self.color_for_tag[tag]

    def __draw_bbox_and_tag(self, image, bbox, tag, bound, meta, deferred_tags):
        color = self.__get_color_for_tag(tag)
        self.__draw_bbox(image, bbox, color)

        if tag:
            left = bound[0][0]
            top = bound[0][1]

            y = top - 15 if top - 15 > 15 else top + 15

            deferred_tags.append({
                "text": tag,
                "origin": (left, y),
                "color": color,
                "meta": meta
            })

    def __random_color(self, value=1.0, saturation=0.51):
        i_max = int(np.random.random() * 3)
        i_mid = (i_max + int(np.random.random() * 2) + 1) % 3
        v_max = int(255 * value)
        v_min = int(v_max * saturation)
        v_mid = int(v_min + int(np.random.random() * (v_max - v_min)))
        color = [v_min, v_min, v_min]
        color[i_max] = v_max
        color[i_mid] = v_mid
        return tuple(color)

    def __contrasting_color(self, color, saturation=0.51):
        def contrast(v): return 1 if v < 128 else 0
        cont = (contrast(color[0]), contrast(color[1]), contrast(color[2]))
        n_hi = cont[0] + cont[1] + cont[2]
        if not n_hi:
            return (0, 0, 0)
        n_lo = 3 - n_hi
        coeff = 255 * n_hi / ((n_lo * saturation) + n_hi)
        def desaturate(v): return int(coeff) if v else int(saturation * coeff)
        return (desaturate(cont[0]), desaturate(cont[1]), desaturate(cont[2]))

    def __draw_polygon(self, image, polygon, color, fill_alpha=0.4, thickness=2, shift=8):
        width, height = image.shape[1], image.shape[0]
        thickness = max(1, int(min(width, height) / 200))

        for verts in polygon:
            np_verts = np.multiply(verts, 256)
            fill = image.copy()
            cv2.fillPoly(fill, [np_verts.astype(int)],
                         color, cv2.LINE_4, shift)
            cv2.addWeighted(fill, fill_alpha, image, 1 - fill_alpha, 0, image)
            cv2.polylines(image, [np_verts.astype(int)], True, color,
                          thickness, cv2.LINE_4, shift)

    def __draw_polygon_and_tag(self, image, polygon, tag, bounds, meta, deferred_tags):

        color = self.__get_color_for_tag(tag)

        self.__draw_polygon(image, polygon, color)

        if tag:
            left = bounds["x"]
            top = bounds["y"]
            right = bounds["x"] + bounds["width"]
            FONT_WIDTH = 10
            FONT_HEIGHT = 16

            y = top - FONT_HEIGHT
            x = (left + right - (FONT_WIDTH * len(tag))) // 2
            org = (max(x, 0), max(y, 2 * FONT_HEIGHT))

            deferred_tags.append({
                "text": tag,
                "origin": org,
                "color": color,
                "meta": meta
            })

    def display(self,
                show_bboxes: bool = False,
                bbox_constraints: Constraints = None,
                show_polygons: bool = False,
                limit: Union[int, object] = None,
                polygon_constraints: Constraints = None,
                polygon_tag_key: str = "_label",
                polygon_tag_format: str = "{}") -> None:
        """
        **Display images with annotations**

        Args:
            show_bboxes: bool, optional
                Show bounding boxes, by default False
            bbox_constraints: Constraints, optional
                Constraints for bounding boxes, by default None
            show_polygons: bool, optional
                Show polygons, by default False
            limit: Union[int, object], optional
                Number of images to display, by default None
            polygon_constraints: Constraints, optional
                Constraints for polygons, by default None
            polygon_tag_key: str, optional
                Key for the polygon tag, by default "_label"
            polygon_tag_format: str, optional
                Format for the polygon tag, by default "{}"
        """
        if not limit:
            limit = self.display_limit
            if self.display_limit < len(self.images_ids):
                print("Showing only first", limit, "results.")

        if polygon_constraints:
            show_polygons = True
            self.images_polygons = {}

        for i in range(len(self.images_ids)):

            if limit == 0:
                break
            limit -= 1

            uniqueid = str(self.images_ids[i])
            image = self.get_image_by_index(i)

            # Just decode the image from buffer
            nparr = np.frombuffer(image, dtype=np.uint8)
            image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

            deferred_tags = []
            if show_bboxes:
                if not str(uniqueid) in self.images_bboxes:
                    self.__retrieve_bounding_boxes(i, bbox_constraints)

                bboxes = self.images_bboxes[uniqueid]["bboxes"]
                tags = self.images_bboxes[uniqueid]["tags"]
                meta = self.images_bboxes[uniqueid]["meta"]
                bounds = self.images_bboxes[uniqueid]["bounds"]

                # Draw a rectangle around the faces
                for bi in range(len(bboxes)):
                    self.__draw_bbox_and_tag(
                        image, bboxes[bi], tags[bi], bounds[bi], meta[bi], deferred_tags)

            if show_polygons:
                if str(uniqueid) not in self.images_polygons:
                    self.__retrieve_polygons(
                        i, polygon_constraints, polygon_tag_key, polygon_tag_format)

                bounds = self.images_polygons[uniqueid]["bounds"] if uniqueid in self.images_polygons else [
                ]
                polygons = self.images_polygons[uniqueid]["polygons"] if uniqueid in self.images_polygons else [
                ]
                tags = self.images_polygons[uniqueid]["tags"] if uniqueid in self.images_polygons else [
                ]
                meta = self.images_polygons[uniqueid]["meta"] if uniqueid in self.images_polygons else [
                ]

                for pi in range(len(polygons)):
                    self.__draw_polygon_and_tag(image, polygons[pi], tags[pi] if pi < len(
                        tags) else None, bounds[pi], meta[pi], deferred_tags)

            for ovr in self.overlays:
                if "polygons" in ovr:
                    self.__draw_polygon(
                        image, ovr["polygons"], ovr["color"], ovr["alpha"])
                elif "bbox" in ovr:
                    self.__draw_bbox(image, ovr["bbox"], ovr["color"])
                elif "text" in ovr:
                    deferred_tags.append(ovr)

            for tag in deferred_tags:
                self.__draw_text_with_shadow(
                    image, tag["text"], tag["origin"], tag["color"], meta=tag["meta"] if "meta" in tag else None)

            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            fig1, ax1 = plt.subplots()
            plt.imshow(image)

    def get_props_names(self) -> List[str]:
        """
        **Get the names of the properties that apply to images**

        Returns:
            properties (List[str]): The names of the properties of the images
        """
        dbutils = Utils.Utils(self.db_connector)
        schema = dbutils.get_schema()

        try:
            dictio = schema["entities"]["classes"]["_Image"]["properties"]
            props_array = [key for key, val in dictio.items()]
        except:
            props_array = []
            print("Cannot retrieve properties")

        return props_array

    def get_properties(self, prop_list: Iterable[str] = []) -> Dict[str, Any]:
        """
        **Get the properties of the images**

        Args:
            prop_list (List[str], optional): The list of properties to retrieve. Defaults to [].

            Returns:
                property_values Dict[str, Any]: The properties of the images
        """
        if len(prop_list) == 0:
            return {}

        return_dictionary = {}

        try:
            for uniqueid in self.images_ids:

                query = [{
                    "FindImage": {
                        "_ref": 1,
                        "constraints": {
                            self.img_id_prop: ["==", uniqueid]
                        },
                        "blobs": False,
                        "results": {
                            "list": prop_list
                        }
                    }
                }]

                res, images = self.db_connector.query(query)

                return_dictionary[str(
                    uniqueid)] = res[0]["FindImage"]["entities"][0]
        except:
            print("Cannot retrieved properties")

        return return_dictionary



from aperturedb import CSVParser
import logging

logger = logging.getLogger(__name__)
# we need to update conditionally; if_not_found only works for add.
# additionally there are two metrics in use: selection critera
# and update critera.
# find_<prop> is the column form to bind


class SingleEntityUpdateDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB General CSV Parser for Adding and Updating Properties in an Entity**


      Update an Entity to the data in the CSV
      What this means is:
      - If it doesn't exist, add it.
      - If it exists, update the properties.



       This class utilizes 2 conditionals
       - normal constraint_ to select the element
       - a series of updateif_ to determine if an update is necessary

       Conditionals:
         updateif>_prop - updates if the databset value > csv value
         updateif<_prop - updates if the database value < csv value
         updateif!_prop - updates if the database value is != csv value

    :::note
    Is backed by a CSV file with the following columns (format optional):

        ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``gs_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
        ..
    :::

    Example CSV file::

        filename,id,label,constraint_id,format,dataset_ver,updateif>_dataset_ver,gen_blobsha1_sha
        /home/user/file1.jpg,321423532,dog,321423532,jpg,2,2,
        /home/user/file2.jpg,42342522,cat,42342522,png,2,2,
        ...

    Example usage:

    ```python
        data = ImageForceNewestDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```


    :::info
    In the above example, the constraint_id ensures that an Image with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """
    UPDATE_CONSTRAINT_PREFIX = "updateif_"

    def __init__(self, entity_class: str, filename: str, **kwargs):

        if entity_class in ("Image", "Video", "Blob", "BoundingBox", "Connection", "Polygon", "Descriptor", "DescriptorSet", "Frame"):
            self.entity = entity_class
            self.entity_selection = None
        else:
            self.entity = "Entity"
            self.entity_selection = entity_class
        self.keys_set = False
        super().__init__(filename, **kwargs)
        # We do not expect a blob in either query, if a subclass adds blob, the first query will need a blob.
        self.blobs_per_query = [0, 0]
        self.commands_per_query = [1, 1]
        self._setupkeys()

    def _setupkeys(self):
        if not self.keys_set:
            self.keys_set = True
            self.props_keys = [x for x in self.header[1:]
                               if not (x.startswith(CSVParser.CONSTRAINTS_PREFIX)
                                       or x.startswith(SingleEntityUpdateDataCSV.UPDATE_CONSTRAINT_PREFIX))]
            self.constraints_keys = [x for x in self.header[1:]
                                     if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
            self.search_keys = [x for x in self.header[1:]
                                if x.startswith(SingleEntityUpdateDataCSV.UPDATE_CONSTRAINT_PREFIX)]

    def getitem(self, idx):
        idx = self.df.index.start + idx
        query_set = []

        self.constraint_keyword = "if_not_found"
        self.command = "Add" + self.entity
        selector = {} if self.entity_selection is None else {
            "class": self.entity_selection}
        entity_add = self._basic_command(idx, custom_fields=selector)
        condition_add_failed = {"results": {0: {"status": ["==", 2]}}}
        self.command = "Update" + self.entity
        update_constraints = self.parse_constraints(idx)
        search_constraints = self.parse_other_constraint(
            SingleEntityUpdateDataCSV.UPDATE_CONSTRAINT_PREFIX, self.search_keys, idx)
        update_constraints.update(search_constraints)
        properties = self.parse_properties(idx)
        self.constraint_keyword = "constraints"
        selector = {} if self.entity_selection is None else {
            "with_class": self.entity_selection}
        entity_update = self._parsed_command(
            idx, selector, update_constraints, properties)
        query_set.append(entity_add)
        query_set.append([condition_add_failed, entity_update])

        if hasattr(self, "modify_item") and callable(self.modify_item):
            query_set = self.modify_item(query_set, idx)

#        raise Exception("State = " +str(query_set))
        return [query_set], []

    def validate(self):
        self._setupkeys()
        valid = True
        if not self.use_dask:
            if len(self.constraints_keys) < 1:
                logger.error("Cannot add/update " +
                             self.entity + "; no constraint keys")
                valid = False
            if valid and len(self.search_keys) < 1:
                logger.error("Cannot update " + self.entity +
                             "; no update constraint keys")
                valid = False
        return valid


class EntityUpdatDataCSV(SingleEntityUpdateDataCSV):
    def __init__(self, entity_type, filename, df=None, use_dask=False):
        super().__init__("Entity", filename, df, use_dask)
        self.entity_type = entity_type
        # Add had blob and update has blob.
        self.blobs_per_query = [0, 0]

    def modify_item(self, query_set, idx):
        query_set[0]["AddEntity"]["class"] = self.entity_type
        return query_set



import logging

logger = logging.getLogger(__name__)

error_message = """
Facenet transformer requires facenet-pytorch and torch
Install with: pip install aperturedb[complete]
Alternatively, install with: "pip install facenet-pytorch torch" in the same
venv as aperturedb.
"""

try:
    from facenet_pytorch import MTCNN, InceptionResnetV1
    import torch
except ImportError:
    logger.critical(error_message)
    exit(1)

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

# If required, create a face detection pipeline using MTCNN:
mtcnn = MTCNN(image_size=96, margin=0, device=device)

# Create an inception resnet (in eval mode):
resnet = InceptionResnetV1(pretrained='vggface2', device=device).eval()

errors = 0


def generate_embedding(img):
    global errors
    # Get cropped and prewhitened image tensor
    img_cropped = mtcnn(img)
    if img_cropped is not None:
        # Calculate embedding (unsqueeze to add batch dimension)
        img_embedding = resnet(img_cropped.unsqueeze(0).to(device))
    else:
        img_embedding = torch.FloatTensor(1, 512).to(device)
        errors += 1

    return img_embedding



from aperturedb.Subscriptable import Subscriptable
from aperturedb.transformers.transformer import Transformer
from .clip import generate_embedding, descriptor_set
from aperturedb.Utils import create_connector, Utils


class CLIPPyTorchEmbeddings(Transformer):
    """
    Generates the embeddings for the images using the CLIP Pytorch model.
    https://github.com/openai/CLIP
    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        """
        Args:
            data: Subscriptable object
            search_set_name: Name of the [descriptorset](/query_language/Reference/descriptor_commands/desc_commands/AddDescriptor) to use for the search.
        """
        super().__init__(data)
        self.search_set_name = kwargs.get(
            "search_set_name", descriptor_set)

        # Let's sample some data to figure out the descriptorset we need.
        if len(self._add_image_index) > 0:
            sample = generate_embedding(self.data[0][1][0])
            utils = Utils(create_connector())
            utils.add_descriptorset(self.search_set_name, dim=len(sample) // 4)

    def getitem(self, subscript):
        x = self.data[subscript]

        for ic in self._add_image_index:
            serialized = generate_embedding(x[1][ic])
            x[1].append(serialized)
            x[0].append(
                {
                    "AddDescriptor": {
                        "set": self.search_set_name,
                        "connect": {
                            "ref": x[0][ic]["AddImage"]["_ref"]
                        }
                    }
                })
        return x



import math
import numpy as np
import cv2
import logging

from aperturedb import Images

from torch.utils import data


logger = logging.getLogger(__name__)


class ApertureDBDataset(data.Dataset):
    """
    This class implements a PyTorch Dataset for ApertureDB.
    It is used to load images from ApertureDB into a PyTorch model.
    It can be initialized with a query that will be used to retrieve
    the images from ApertureDB.
    """

    def __init__(self, db, query, label_prop=None, batch_size=1):

        self.db = db.create_new_connection()
        self.query = query
        self.find_image_idx = None
        self.total_elements = 0
        self.batch_size     = batch_size
        self.batch_images   = []
        self.batch_start    = 0
        self.batch_end      = 0
        self.label_prop     = label_prop

        for i in range(len(query)):

            name = list(query[i].keys())[0]
            if name == "FindImage":
                self.find_image_idx = i

        if self.find_image_idx is None:
            logger.error(
                "Query error. The query must contain one FindImage command")
            raise Exception('Query Error')

        if not "results" in self.query[self.find_image_idx]["FindImage"]:
            self.query[self.find_image_idx]["FindImage"]["results"] = {}

        self.query[self.find_image_idx]["FindImage"]["batch"] = {}
        self.query[self.find_image_idx]["FindImage"]["blobs"] = True

        try:
            r, b = self.db.query(self.query)
            batch = r[self.find_image_idx]["FindImage"]["batch"]
            self.total_elements = batch["total_elements"]
        except:
            logger.error(
                f"Query error: {self.query} {self.db.get_last_response_str()}")
            raise

    def __getitem__(self, index):

        if index >= self.total_elements:
            raise StopIteration

        if not self.is_in_range(index):
            self.get_batch(index)

        idx = index % self.batch_size
        img   = self.batch_images[idx]
        label = self.batch_labels[idx]

        nparr = np.frombuffer(img, dtype=np.uint8)
        img   = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        img   = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return img, label

    def __len__(self):

        return self.total_elements

    def is_in_range(self, index):

        if index >= self.batch_start and index < self.batch_end:
            return True

        return False

    def get_batch(self, index):

        total_batches = math.ceil(self.total_elements / self.batch_size)
        batch_idx     = math.floor(index / self.batch_size)

        if batch_idx >= total_batches:
            raise Exception("Index out of range")

        query  = self.query
        qbatch = query[self.find_image_idx]["FindImage"]["batch"]
        qbatch["batch_size"] = self.batch_size
        qbatch["batch_id"]   = batch_idx

        query[self.find_image_idx]["FindImage"]["batch"] = qbatch

        try:

            # This is to handle potential issues with
            # disconnection/timeout and SSL context on multiprocessing
            connection_ok = False
            try:
                r, b = self.db.query(query)
                connection_ok = True
            except:
                # Connection failed, we retry just once to re-connect
                self.db = self.db.create_new_connection()

            if not connection_ok:
                # Connection failed, we have reconnected, we try again.
                r, b = self.db.query(query)

            if len(b) == 0:
                logger.error(f"index: {index}")
                raise Exception("No results returned from ApertureDB")

            self.batch_images = b
            self.batch_start  = self.batch_size * batch_idx
            self.batch_end    = self.batch_start + len(b)

            if self.label_prop:
                entities = r[self.find_image_idx]["FindImage"]["entities"]
                self.batch_labels = [l[self.label_prop] for l in entities]
            else:
                self.batch_labels = ["none" for l in range(len(b))]
        except:
            logger.error(f"Query error: {self.db.get_last_response_str()}")
            raise



from __future__ import annotations
from typing import Callable, Dict, List, Tuple, Optional
from aperturedb import Parallelizer
import numpy as np
import json
import logging
import math


from aperturedb.DaskManager import DaskManager
from aperturedb.Connector import Connector
from aperturedb.types import Commands, Blobs, CommandResponses

logger = logging.getLogger(__name__)


def execute_batch(q: Commands, blobs: Blobs, db: Connector,
                  success_statuses: list[int] = [0],
                  response_handler: Optional[Callable] = None, commands_per_query: int = 1, blobs_per_query: int = 0,
                  strict_response_validation: bool = False) -> Tuple[int, CommandResponses, Blobs]:
    """
    Execute a batch of queries, doing useful logging around it.
    Calls the response handler if provided.
    This should be used (without the parallel machinery) instead of db.query to keep the response handling consistent, better logging, etc.

    Args:
        q (Commands): List of commands to execute.
        blobs (Blobs): List of blobs to send.
        db (Connector): The database connector.
        success_statuses (list[int], optional): The list of success statuses. Defaults to [0].
        response_handler (Callable, optional): The response handler. Defaults to None.
        commands_per_query (int, optional): The number of commands per query. Defaults to 1.
        blobs_per_query (int, optional): The number of blobs per query. Defaults to 0.
        strict_response_validation (bool, optional): Whether to strictly validate the response. Defaults to False.

    Returns:
        int: The result code.
            - 0 : if all commands succeeded
            - 1 : if there was -1 in the response
            - 2 : For any other code.
        CommandResponses: The response.
        Blobs: The blobs.
    """
    result = 0
    logger.debug(f"Query={q}")
    r, b = db.query(q, blobs)
    logger.debug(f"Response={r}")

    if db.last_query_ok():
        if response_handler is not None:
            # We could potentially always call this handler function
            # and let the user deal with the error cases.
            blobs_returned = 0
            for i in range(math.ceil(len(q) / commands_per_query)):
                start = i * commands_per_query
                end = start + commands_per_query
                blobs_start = i * blobs_per_query
                blobs_end = blobs_start + blobs_per_query

                b_count = 0
                for req, resp in zip(q[start:end], r[start:end]):
                    for k in req:
                        # Ref to https://docs.aperturedata.io/query_language/Reference/shared_command_parameters/blobs
                        blobs_where_default_true = \
                            k in ["FindImage", "FindBlob", "FindVideo"] and (
                                "blobs" not in req[k] or req[k]["blobs"])
                        blobs_where_default_false = \
                            k in [
                                "FindDescriptor", "FindBoundingBox"] and "blobs" in req[k] and req[k]["blobs"]
                        if blobs_where_default_true or blobs_where_default_false:
                            count = resp[k]["returned"]
                            b_count += count

                try:
                    # The returned blobs need to be sliced to match the
                    # returned entities per command in query.
                    response_handler(
                        q[start:end],
                        blobs[blobs_start:blobs_end],
                        r[start:end],
                        b[blobs_returned:blobs_returned + b_count] if len(b) >= blobs_returned + b_count else None)
                except BaseException as e:
                    logger.exception(e)
                    if strict_response_validation:
                        raise e
                blobs_returned += b_count
    else:
        # Transaction failed entirely.
        logger.error(f"Failed query = {q} with response = {r}")
        result = 1

    statuses = {}
    if isinstance(r, dict):
        statuses[r['status']] = [r]
    elif isinstance(r, list):
        # add each result to a list of the responses, keyed by the response
        # code.
        [statuses.setdefault(result[cmd]['status'], []).append(result)
         for result in r for cmd in result]
    else:
        logger.error("Response in unexpected format")
        result = 1

    # last_query_ok means result status >= 0
    if result != 1:
        warn_list = []
        for status, results in statuses.items():
            if status not in success_statuses:
                for wr in results:
                    warn_list.append(wr)
        if len(warn_list) != 0:
            logger.warning(
                f"Partial errors:\r\n{json.dumps(q)}\r\n{json.dumps(warn_list)}")
            result = 2

    return result, r, b


class ParallelQuery(Parallelizer.Parallelizer):
    """
    **Parallel and Batch Querier for ApertureDB**

    This class provides the abstraction for partitioning data into batches,
    so that they may be processed using different threads.

    Args:
        db (Connector): The database connector.
        dry_run (bool, optional): Whether to run in dry run mode. Defaults to False.
    """

    # 0 is success, 2 is object exists
    success_statuses = [0, 2]

    @classmethod
    def setSuccessStatus(cls, statuses: list[int]):
        cls.success_statuses = statuses

    @classmethod
    def getSuccessStatus(cls):
        return cls.success_statuses

    def __init__(self, db: Connector, dry_run: bool = False):

        super().__init__()

        self.db = db.create_new_connection()

        self.dry_run = dry_run

        self.type = "query"

        self.responses = []

        self.commands_per_query = 1
        self.blobs_per_query = 0
        self.daskManager = None
        self.batch_command = execute_batch

    def generate_batch(self, data: List[Tuple[Commands, Blobs]]) -> Tuple[Commands, Blobs]:
        """
        Here we flatten the individual queries to run them as
        a single query in a batch
        We also update the _ref values and connections refs.

        Args:
            data (list[tuple[Query, Blobs]]): The data to be batched.  Each tuple contains a list of commands and a list of blobs.

        Returns:
            commands (Commands): The batched commands.
            blobs (Blobs): The batched blobs.
        """
        def update_refs(batched_commands):
            updates = {}
            for i, cmd in enumerate(batched_commands):
                if isinstance(cmd, list):
                    # Only parallel queries will work.
                    break
                values = cmd[list(cmd.keys())[0]]
                if "_ref" in values:
                    updates[values["_ref"]] = i + 1
                    values["_ref"] = i + 1
                    assert values["_ref"] < 100000
                if "image_ref" in values:
                    values["image_ref"] = updates[values["image_ref"]]
                if "video_ref" in values:
                    values["video_ref"] = updates[values["video_ref"]]
                if "is_connected_to" in values and "_ref" in values["is_connected_to"]:
                    values["is_connected_to"]["_ref"] = updates[values["is_connected_to"]["_ref"]]
                if "connect" in values and "ref" in values["connect"]:
                    values["connect"]["ref"] = updates[values["connect"]["ref"]]
                if "src" in values:
                    values["src"] = updates[values["src"]]
                if "dst" in values:
                    values["dst"] = updates[values["dst"]]
            return batched_commands

        q = update_refs([cmd for query in data for cmd in query[0]])
        blobs = [blob for query in data for blob in query[1]]

        return q, blobs

    def call_response_handler(self, q: Commands, blobs: Blobs, r: CommandResponses, b: Blobs):
        try:
            self.generator.response_handler(q, blobs, r, b)
        except BaseException as e:
            logger.exception(e)

    def do_batch(self, db: Connector, data: List[Tuple[Commands, Blobs]]) -> None:
        """
        Executes batch of queries and blobs in the database.

        Args:
            db (Connector): The database connector.
            data (list[tuple[Commands, Blobs]]): The data to be batched.  Each tuple contains a list of commands and a list of blobs.

        It also provides a way for invoking a user defined function to handle the
        responses of each of the queries executed. This function can be used to process
        the responses from each of the corresponding queries in [Parallelizer](/python_sdk/parallel_exec/Parallelizer)
        It will be called once per query, and it needs to have 4 parameters:
        - requests
        - input_blobs
        - responses
        - output_blobs
        Example usage:
        ``` python
            class MyQueries(QueryGenerator):
                def process_responses(requests, input_blobs, responses, output_blobs):
                    self.requests.extend(requests)
                    self.responses.extend(responses)
            loader = ParallelLoader(self.db)
            generator = MyQueries()
            loader.ingest(generator)
        ```
        """

        q, blobs = self.generate_batch(data)

        query_time = 0
        worker_stats = {}
        if not self.dry_run:
            response_handler = None
            strict_response_validation = False
            if hasattr(self.generator, "response_handler") and callable(self.generator.response_handler):
                response_handler = self.generator.response_handler
            if hasattr(self.generator, "strict_response_validation") and isinstance(self.generator.strict_response_validation, bool):
                strict_response_validation = self.generator.strict_response_validation
            result, r, b = self.batch_command(
                q,
                blobs,
                db,
                ParallelQuery.success_statuses,
                response_handler,
                self.commands_per_query,
                self.blobs_per_query,
                strict_response_validation=strict_response_validation)
            if result == 0:
                query_time = db.get_last_query_time()
                worker_stats["succeeded_commands"] = len(q)
                worker_stats["succeeded_queries"] = len(data)
                worker_stats["objects_existed"] = sum(
                    [v['status'] == 2 for i in r for k, v in i.items()])
            elif result == 1:
                self.error_counter += 1
                worker_stats["succeeded_queries"] = 0
                worker_stats["succeeded_commands"] = 0
                worker_stats["objects_existed"] = 0
            elif result == 2:
                # with result 2, some queries might have failed.
                def filter_per_group(group):
                    return group.items() if isinstance(group, dict) else {}
                worker_stats["succeeded_commands"] = sum(
                    [v['status'] == 0 for i in r for k, v in filter_per_group(i)])
                worker_stats["objects_existed"] = sum(
                    [v['status'] == 2 for i in r for k, v in filter_per_group(i)])
                sq = 0
                for i in range(0, len(r), self.commands_per_query):
                    if all([v['status'] == 0 for j in r[i:i + self.commands_per_query] for k, v in filter_per_group(j)]):
                        sq += 1
                worker_stats["succeeded_queries"] = sq
        else:
            query_time = 1

        # append is thread-safe
        self.times_arr.append(query_time)
        self.actual_stats.append(worker_stats)

    def worker(self, thid: int, generator, start: int, end: int):
        # A new connection will be created for each thread
        db = self.db.create_new_connection()

        total_batches = (end - start) // self.batchsize

        if (end - start) % self.batchsize > 0:
            total_batches += 1

        logger.info(f"Worker {thid} executing {total_batches} batches")
        for i in range(total_batches):

            batch_start = start + i * self.batchsize
            batch_end = min(batch_start + self.batchsize, end)

            try:
                self.do_batch(db, generator[batch_start:batch_end])
            except Exception as e:
                logger.exception(e)
                logger.warning(
                    f"Worker {thid} failed to execute batch {i}: [{batch_start},{batch_end}]")
                self.error_counter += 1

            if thid == 0 and self.stats:
                self.pb.update((i + 1) / total_batches)

    def get_objects_existed(self) -> int:
        return sum([stat["objects_existed"]
                    for stat in self.actual_stats])

    def get_succeeded_queries(self) -> int:
        return sum([stat["succeeded_queries"]
                    for stat in self.actual_stats])

    def get_succeeded_commands(self) -> int:
        return sum([stat["succeeded_commands"]
                    for stat in self.actual_stats])

    def query(self, generator, batchsize: int = 1, numthreads: int = 4, stats: bool = False) -> None:
        """
        This function takes as input the data to be executed in specified number of threads.
        The generator yields a tuple : (array of commands, array of blobs)
        Args:
            generator (_type_): The class that generates the queries to be executed.
            batchsize (int, optional): Number of queries per transaction. Defaults to 1.
            numthreads (int, optional): Number of parallel workers. Defaults to 4.
            stats (bool, optional): Show statistics at end of ingestion. Defaults to False.
        """

        use_dask = hasattr(generator, "use_dask") and generator.use_dask
        if use_dask:
            self._reset(batchsize=batchsize, numthreads=numthreads)
            self.daskmanager = DaskManager(num_workers=numthreads)

        if hasattr(self, "query_setup"):
            self.query_setup(generator)

        if use_dask:
            results, self.total_actions_time = self.daskmanager.run(
                self.__class__, self.db, generator, batchsize, stats=stats)
            self.actual_stats = []
            for result in results:
                if result is not None:
                    self.times_arr.extend(result.times_arr)
                    self.error_counter += result.error_counter
                    self.actual_stats.append(
                        {"succeeded_queries": result.succeeded_queries,
                         "succeeded_commands": result.succeeded_commands,
                         "objects_existed": result.objects_existed})
            self.total_actions = len(generator.df)

            if stats:
                self.print_stats()
        else:
            # allow subclass to do verification
            if issubclass(type(self), ParallelQuery) and hasattr(self, 'verify_generator') and callable(self.verify_generator):
                self.verify_generator(generator)
            elif len(generator) > 0:
                if isinstance(generator[0], tuple) and isinstance(generator[0][0], list):
                    # if len(generator[0]) > 0:
                    #
                    #  Not applicable to old style loaders.
                    self.commands_per_query = len(generator[0][0])
                    if len(generator[0][1]):
                        self.blobs_per_query = len(generator[0][1])
                else:
                    logger.error(
                        f"Could not determine query structure from:\n{generator[0]}")
                    logger.error(type(generator[0]))
            logger.info(
                f"Commands per query = {self.commands_per_query}, Blobs per query = {self.blobs_per_query}"
            )
            self.run(generator, batchsize, numthreads, stats)

    def print_stats(self) -> None:

        times = np.array(self.times_arr)
        total_queries_exec = len(times)

        print("============ ApertureDB Parallel Query Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 0 if mean == 0 else 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

        print("=========================================================")

    def debug_sample(self, **kwargs) -> None:
        """
        Sample the data to be ingested for debugging purposes.
        """
        if "sample_count" in kwargs:
            sample_count = kwargs["sample_count"]
        else:
            sample_count = len(self.generator)

        if sample_count > 0:
            print("Sample of data to be ingested:")
            for i in range(sample_count):
                print(self.generator[i])



from typing import List, Tuple
from torch.utils.data import Dataset
from aperturedb.Subscriptable import Subscriptable


class PyTorchData(Subscriptable):
    """
    **Class to wrap around a Dataset retrieved from [PyTorch datasets](https://pytorch.org/vision/0.15/datasets.html)**

    The dataset in this case can be iterated over.
    So the only thing that needs to be implemented is generate_query,
    which takes an index and returns a query.

    :::note
    This class should be subclassed with a specific (custom) implementation of generate_query().
    :::

    Example subclass: [CocoDataPyTorch](https://github.com/aperture-data/aperturedb-python/blob/develop/examples/CocoDataPyTorch.py)

    """

    def __init__(self, dataset: Dataset) -> None:
        self.loaded_dataset = [t for t in dataset]

    def getitem(self, idx: int):
        return self.generate_query(idx)

    def __len__(self):
        return len(self.loaded_dataset)

    def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
        """
        **Takes information from one atomic record from the Data and converts it to Query for apertureDB**

        Args:
            idx (int): index of the record in collection.

        Raises:
            Exception: _description_

        Returns:
            Tuple[List[dict], List[bytes]]: A pair of list of commands and optional list of blobs to go with them.
        """
        raise Exception("To be implemented by subclass")



import socket
import time
import logging
import ssl
import dbinfo

from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest

logger = logging.getLogger(__name__)


class TestSession():
    """
    These check operation of the Connector Session
    and the error handling with the Network layer
    """

    def test_sessionRenew(self, db: Connector):
        """
        Verifies that Session renew works
        """

        # force session token expiry
        db.shared_data.session.session_token_ttl = 1
        logger.debug("Connected? {0}".format(
            "yes" if db.connected else "no"))
        logger.debug(
            "Session valid? {0}".format(
                "yes" if db.shared_data.session.valid() else "no"))
        logger.debug("Valid length: {0}".format(
            db.shared_data.session.session_token_ttl))
        time.sleep(2)
        query = [{
            "FindImage": {
                "results": {
                    "limit": 5
                }
            }
        }]
        responses, blobs = db.query(query)
        logger.debug(responses)
        logger.debug("Valid : {0}".format(
            db.shared_data.session.valid()))
        assert db.shared_data.session.valid() == True

    def test_SSL_error_on_query(self, monkeypatch):
        db = Connector(
            dbinfo.DB_TCP_HOST,
            dbinfo.DB_TCP_PORT,
            dbinfo.DB_USER,
            dbinfo.DB_PASSWORD,
            retry_max_attempts=3,
            retry_interval_seconds=0)
        db.query([{"FindImage": {"results": {"limit": 5}}}])
        original_send_msg = db._send_msg
        count = 0

        # This will raise an SSL error on the first call to _send_msg (which is an refresh token)
        # The later calls will succeed.
        # When we try to renew the session, that attempts to take a lock on the session.
        # And this is the hang.
        def mock_send_msg(msg):
            nonlocal count
            count += 1
            logger.debug(count)
            if count == 1:
                raise ssl.SSLError("SSL Error")

            result  = original_send_msg(msg)
            logger.debug(result)
            return result

        monkeypatch.setattr(db, "_send_msg", lambda x: mock_send_msg(x))

        # force session token expiry
        db.shared_data.session.session_token_ttl = 1
        time.sleep(2)
        query = [{
            "FindImage": {
                "results": {
                    "limit": 5
                }
            }
        }]
        responses, blobs = db.query(query)
        logging.debug(responses)
        assert db.shared_data.session.valid() == True

    def test_socket_connect_error_initial(self, monkeypatch):
        connect_attempts = 0

        def mock_connect(host, port):
            nonlocal connect_attempts
            connect_attempts += 1
            raise ConnectionRefusedError("Connection Refused")
        monkeypatch.setattr(socket.socket, "connect",
                            lambda h, p: mock_connect(h, p))

        # Create new db connection.

        new_db = Connector(
            dbinfo.DB_TCP_HOST,
            dbinfo.DB_TCP_PORT,
            dbinfo.DB_USER,
            dbinfo.DB_PASSWORD,
            retry_max_attempts=3,
            retry_interval_seconds=0)
        try:
            new_db.query([{"FindImage": {"results": {"limit": 5}}}])
        except Exception as e:
            # Check the exception is not an obscure one.
            assert e.args[0] == "Authentication failed:"

        # Check that we tried to connect 3 times.
        assert connect_attempts == 3

    def test_socket_send_error_initial(self, monkeypatch):
        send_attempts = 0

        def mock_send(x, buff):
            nonlocal send_attempts
            send_attempts += 1
            raise socket.error("Connection broke when send")
        monkeypatch.setattr(socket.socket, "send", mock_send)

        # Create new db connection.
        new_db = Connector(
            dbinfo.DB_TCP_HOST,
            dbinfo.DB_TCP_PORT,
            dbinfo.DB_USER,
            dbinfo.DB_PASSWORD,
            retry_max_attempts=3,
            retry_interval_seconds=0)
        try:
            new_db.query([{"FindImage": {"results": {"limit": 5}}}])
        except Exception as e:
            # Check the exception is not an obscure one.
            assert e.args[0] == "Authentication failed:"

        # Check that we tried to send 5 (connect hello:2) + query:3) times.
        assert send_attempts == 5

    def test_socket_recv_error_initial(self, monkeypatch):
        connect_attempts = 0

        def mock_recv(x, buff):
            nonlocal connect_attempts
            connect_attempts += 1
            # raise socket.error("Connection broke when recv")
            raise ConnectionResetError("Connection broke when recv")
        monkeypatch.setattr(socket.socket, "recv", mock_recv)

        # Create new db connection.

        new_db = Connector(
            dbinfo.DB_TCP_HOST,
            dbinfo.DB_TCP_PORT,
            dbinfo.DB_USER,
            dbinfo.DB_PASSWORD,
            retry_max_attempts=3,
            retry_interval_seconds=0)
        try:
            new_db.query([{"FindImage": {"results": {"limit": 5}}}])
        except Exception as e:
            # Check the exception is not an obscure one.
            assert e.args[0] == "Authentication failed:"

        # Check that we tried to connect 3 times.
        assert connect_attempts == 3

    def test_con_close_on_send_query(self, db: Connector, monkeypatch):
        if not isinstance(db, ConnectorRest):
            original_send_msg = db._send_msg
            count = 0

            def mock_send_msg(msg):
                nonlocal count
                count += 1
                logger.debug(count)
                if count < 2:
                    # Fail for the firs 2 attempts
                    db.conn.close()

                result  = original_send_msg(msg)
                logger.debug(result)
                return result

            monkeypatch.setattr(db, "_send_msg", mock_send_msg)

            query = [{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }]
            response, blobs = db.query(query)
            assert(response[0]["FindImage"]["status"] == 0)
            assert count == 3

    def test_con_close_on_recv_query(self, db: Connector, monkeypatch):
        if not isinstance(db, ConnectorRest):
            original_recv_msg = db._recv_msg
            count = 0

            def mock_recv_msg():
                nonlocal count
                count += 1
                logger.debug(count)
                if count < 2:
                    # Fail for the firs 2 attempts
                    db.conn.close()

                result  = original_recv_msg()
                logger.debug(result)
                return result

            monkeypatch.setattr(db, "_recv_msg", mock_recv_msg)

            query = [{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }]
            response, blobs = db.query(query)
            assert(response[0]["FindImage"]["status"] == 0)
            assert count == 3



from aperturedb import CSVParser
import logging
import hashlib  # for sha1
from datetime import datetime

logger = logging.getLogger(__name__)


class BlobNewestDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB General CSV Parser for Maintaining a Blob set with changing blob data.**

      Use this only when deleting entities is ok.

      Update an Entity which has an associated blob to the data in the CSV
      What this means is:
      - If it doesn't exist, add it.
      - If it exists and the blob hasn't changed, update it.
      - If it exists and the blob has changed, delete and re-add it.

       This means if these elements are part of a graph where they are linked by connections
        these connections will be need to be regenerated afterwards.

       Additionally note that this loader does not notice if a entity is removed.
       - If you wish to clean up removed entities, an increasing load id will cause
         all existing entities to be updated to the newest, and you can delete
         using the old load id.

       This class utilizes 3 conditionals
       - normal constraint_ to select the element
       - a series of updateif_ to determine if an update is necessary
       - one or more prop_ and the associated updateif blob conditionals
         to determine if a update or an delete/add is appropriate

       Generated fields
        Format is: gen_<type>_name

        type:
        - blobsha1 - the sha1 for the blob is calculated
        - blobsize - the length in bytes of the blob is calculate
        - insertdate - ISO Format of date ( this will always change! )

        the result is then used to identify if a blob has changed.

        The generated fields are to be left empty in the CSV input.

        Summary
         This requires a constraint to be able to check if an id exists, and a generated prop
          to be able to detect if the blob matches. It will ensure only one entity exists with
          the constraints and matching the managed blob constraints.

    :::note
    Is backed by a CSV file with the following columns (format optional):

        ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

        OR

        ``gs_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
        ...
    :::

    Example CSV file::

        filename,id,label,constraint_id,format,dataset_ver,updateif>_dataset_ver,gen_blobsha1_sha
        /home/user/file1.jpg,321423532,dog,321423532,jpg,2,2,
        /home/user/file2.jpg,42342522,cat,42342522,png,2,2,
        ...

    Example usage:

    ``` python

        data = ImageForceNewestDataCSV("/path/to/WorkingImageDataset.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```

    :::info
    In the above example, the constraint_id ensures that an Image with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """
    UPDATE_CONSTRAINT_PREFIX = "updateif_"
    GENERATE_PROP_PREFIX = "gen_"

    def __init__(self, entity_class: str, filename: str, **kwargs):
        self.known_generators = ["blobsize", "blobsha1", "insertdate"]
        self.entity = entity_class
        self.keys_set = False
        super().__init__(filename, **kwargs)
        # these tell the query set code how many blobs and commands to expect in each query
        # query #1 = 1 blob, 1 command.
        # query #2 = 0 blobs, 1 command.
        # query #3 = 1 blob, 2 commands.

        # defintion of how many blobs and commands to expect per query, for ParallelQuerySet.
        self.blobs_per_query = [1, 0, 1]
        self.commands_per_query = [1, 1, 2]
        self._setupkeys()

        self._generated_cache = {}

    def _setupkeys(self):
        if not self.keys_set:
            self.keys_set = True
            self.props_keys = [x for x in self.header[1:]
                               if not (x.startswith(CSVParser.CONSTRAINTS_PREFIX)
                                       or x.startswith(BlobNewestDataCSV.UPDATE_CONSTRAINT_PREFIX)
                                       or x.startswith(BlobNewestDataCSV.GENERATE_PROP_PREFIX))]
            self.generated_keys = [x for x in self.header[1:]
                                   if x.startswith(BlobNewestDataCSV.GENERATE_PROP_PREFIX)]
            self.constraints_keys = [x for x in self.header[1:]
                                     if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
            self.search_keys = [x for x in self.header[1:]
                                if x.startswith(BlobNewestDataCSV.UPDATE_CONSTRAINT_PREFIX)]

    # derived class interface for retrieving blob
    def read_blob(self, idx):
        raise Exception(
            "No Blob Defined for BlobNewestDataCSV ( requires subclass )")

    # creates generated data for an index based on supplied action
    def parse_generated(self, idx, action):
        generated = None
        if action == "blobsize":
            blob = self.read_blob(idx)
            generated = len(blob)
        elif action == "blobsha1":
            blob = self.read_blob(idx)
            generated = hashlib.sha1(blob).hexdigest()
        elif action == "insertdate":
            generated = datetime.now().isoformat()
        else:
            raise Exception(f"Unable to generate data for action {action}")
        return generated

    # filter in or out generated constraints
    def filter_generated_constraints(self, return_generated=False):
        filtered = []
        prefix_len = len(BlobNewestDataCSV.UPDATE_CONSTRAINT_PREFIX)
        for key in self.search_keys:
            key_postfix = key[prefix_len:]
            if "_" in key_postfix:
                (action, *proplist) = key_postfix.split('_')
                if action in self.known_generators:
                    if return_generated:
                        filtered.append(key)
                else:
                    # not a match - verify that this is a known property then.
                    if key_postfix[0] in [">", "<"]:
                        key_postfix = key_postfix[1:]
                    if key_postfix not in self.props_keys:
                        raise Exception(
                            f"Column {key} is a constraint, but {key_postfix} does not match a column, and {action} is not a known generated constraint type")
                    else:
                        if not return_generated:
                            filtered.append(key)
            else:
                if not return_generated:
                    filtered.append(key)
        return filtered

    # create generated constraints for specific index
    # match controls if constrain will be a positive selection (True) or negative(False)
    def create_generated_constraints(self, idx, match=True):
        constraints = {}
        generated_keys = self.filter_generated_constraints(
            return_generated=True)
        prefix_len = len(BlobNewestDataCSV.UPDATE_CONSTRAINT_PREFIX)
        for key in generated_keys:
            (action, *proplist) = key[prefix_len:].split('_')
            prop = '_'.join(proplist)
            op = "==" if match else "!="
            cache_key = f"{idx}_{action}"
            v = None
            if cache_key in self._generated_cache:
                v = self._generated_cache[cache_key]
            else:
                v = self.parse_generated(idx, action)
                # save result
                self._generated_cache[cache_key] = v
            constraints[prop] = [op, v]
        return constraints

    # create generated props for specific index
    def create_generated_props(self, idx):
        prefix_len = len(BlobNewestDataCSV.GENERATE_PROP_PREFIX)
        properties = {}
        blob = None
        for generate in self.generated_keys:
            (action, *proplist) = generate[prefix_len:].split('_')
            prop = '_'.join(proplist)
            cache_key = f"{idx}_{action}"
            v = None
            if cache_key in self._generated_cache:
                v = self._generated_cache[cache_key]
            else:
                v = self.parse_generated(idx, action)
                # save result
                self._generated_cache[cache_key] = v
            properties[prop] = v

        return properties

    def getitem(self, idx):
        idx = self.df.index.start + idx
        query_set = []

        # process is; add if not existing ( # Pt 1 )
        # if existing
        # if blob checks pass ( or there are none ) - update metadata ( Pt 2 )
        # if blob check FAILS
        # delete ( Pt 3 )
        # re-add ( Pt 4 )

        # In terms of the blobs/commands per query -
        # query #1 - add, 1 blob, 1 command.
        # query #2 - update. 0 blobs, 1 command.
        # query # - delete/re-add 1 blob( for re-add )  2 commands.

        self.constraint_keyword = "if_not_found"
        self.command = "Add" + self.entity

        # Part 1
        properties = self.parse_properties(idx)
        constraints = self.parse_constraints(idx)
        # we add generated prop, but do not test generated constraints, as they are used to differentiate between
        # an existing entity.
        gen_props = self.create_generated_props(idx)
        properties.update(gen_props)

        entity_add = self._parsed_command(idx, None, constraints, properties)

        # Part 2
        condition_add_failed = {"results": {0: {"status": ["==", 2]}}}
        self.command = "Update" + self.entity
        update_constraints = self.parse_constraints(idx)
        search_constraints = self.parse_other_constraint(BlobNewestDataCSV.UPDATE_CONSTRAINT_PREFIX,
                                                         self.filter_generated_constraints(), idx)
        # we test generated constraints here as they will stop an update from happening.
        generated_positive_constraints = self.create_generated_constraints(
            idx, match=True)
        update_constraints.update(search_constraints)
        update_constraints.update(generated_positive_constraints)
        properties = self.parse_properties(idx)
        self.constraint_keyword = "constraints"
        entity_update = self._parsed_command(
            idx, None, update_constraints, properties)

        # Part 3
        condition_add_and_update_failed = {"results": {
            0: {"status": ["==", 2]},  # exists
            1: {"count": ["==", 0]},  # but didn't update.
        }}

        self.command = "Delete" + self.entity
        entity_delete = self._parsed_command(idx, None, constraints, None)

        query_set.append(entity_add)
        query_set.append([condition_add_failed, entity_update])
        query_set.append([condition_add_and_update_failed,
                         [entity_delete, entity_add]])

        if hasattr(self, "modify_item") and callable(self.modify_item):
            query_set = self.modify_item(query_set, idx)

        blob = self.read_blob(idx)
        # blobs , 1 for add set, 0 for update set, 1 for delete/add set
        blob_set = [[blob], [], [blob]]

        return [[query_set], [blob_set]]

    def validate(self):
        self._setupkeys()
        valid = True
        if not self.use_dask:
            if len(self.constraints_keys) < 1:
                logger.error("Cannot add/update " +
                             self.entity + "; no constraint keys")
                valid = False
            if valid and len(self.search_keys) < 1:
                logger.error("Cannot update " + self.entity +
                             "; no update constraint keys")
                valid = False
            if len(self.filter_generated_constraints()) < 1:
                logger.error("Cannot differentiate update and reinsert for  " +
                             self.__class__.__name__ + ": no generated constraints for blob comparison")
                valid = False
        return valid



import numpy as np
from aperturedb.ParallelQuery import ParallelQuery
from aperturedb.Query import QueryBuilder

import logging
logger = logging.getLogger(__name__)


class TestLoaderSuccess():
    def assertEqual(self, expected, actual):
        if expected != actual:
            raise AssertionError(
                "Expected {}, got {}".format(expected, actual))

    def test_Loader(self, utils, insert_data_from_csv):
        # Assert that we have a clean slate to begin with.
        assert utils.remove_all_indexes()
        assert utils.remove_all_objects() == True
        # initial load
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/persons-exist-base.adb.csv")
        self.assertEqual(len(data), utils.count_entities("Person"))

        # default configuration does not consider object exists to be a query failure
        def assert_partial(loader, test_data):
            self.assertEqual(len(data) + len(test_data) - loader.get_objects_existed(),
                             utils.count_entities("Person"))
        data, _ = insert_data_from_csv(in_csv_file = "./input/persons-some-exist.adb.csv",
                                       loader_result_lambda = assert_partial)

        # change to disallow object exist to qualify as success.
        old_status = ParallelQuery.getSuccessStatus()
        ParallelQuery.setSuccessStatus([0])

        # Assert that we have a clean slate to begin with.
        assert utils.remove_all_indexes()
        assert utils.remove_all_objects() == True
        # initial load
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/persons-exist-base.adb.csv")
        # default configuration does not consider object exists to be a query
        # failure
        data, _ = insert_data_from_csv(
            in_csv_file = "./input/persons-some-exist.adb.csv",
            expected_error_count = 3,
            loader_result_lambda=assert_partial)

        # reset success status to default
        ParallelQuery.setSuccessStatus(old_status)



from __future__ import annotations
from typing import Any, Callable, List, Tuple
import itertools
import logging

import numpy as np

from aperturedb.ParallelQuery import ParallelQuery
from aperturedb.Connector import Connector

from aperturedb.types import Command, Blob, Commands, Blobs, CommandResponses

logger = logging.getLogger(__name__)

# Turn on to debug constraints ( produces a lot of output )
DEBUG_CONSTRAINTS = True

# removes blobs from a list or tuple for pretty printing


def remove_blobs(item: Any) -> Any:
    if isinstance(item, list):
        item = list(map(remove_blobs, item))
    elif isinstance(item, tuple):
        item = list(map(remove_blobs, item))
    elif isinstance(item, bytes):
        item = "*BLOB*"
    return item


def gen_execute_batch_sets(base_executor, per_batch_response_handler: Callable = None):

    #
    # execute_batch_sets - executes multiple sets of queries with optional constraints on follow on sets
    #  query_set: list of queries that get passed to ParallelLoader's execute_batch
    #  blob_set:  list of blobs ( or list of list of blobs ) that get passed to execute_batch
    #  db: Connection object
    #  success_statutes: list of return values from ApertureDB to be considered 'success'
    #  response_handler: optional function to return after an execution ( not active yet )
    #  commands_per_query : list of how many commands each query has.
    #  blobs_per_query: list of how many blobs each query has.
    #  strict_response_validation: same as execute_batch.
    #
    # if blob_set is None, or an empty list, it is ignored.
    # if blob_set is a list, it will be given to the seed query
    # if blob_set is a list of lists, each inner list will be given to the inner
    #  execution
    #
    def execute_batch_sets(query_set, blob_set, db, success_statuses: list[int] = [0],
                           response_handler: Callable = None, commands_per_query: list[int] = -1,
                           blobs_per_query: list[int] = -1, strict_response_validation: bool = False):

        logger.info("Execute Batch Sets = Batch Size {0}  Comands Per Query {1} Blobs Per Query {2}".format(
            len(query_set), commands_per_query, blobs_per_query))

        batch_size = len(query_set)

        # test query set
        first_element = query_set[0]
        if not isinstance(first_element, list):
            logger.error("First Element not a list: {first_element}")
            raise Exception("Query set must be a list of lists")
        set_total = len(first_element)

        # Check if blobs are a simple array or nested array of blobs
        per_set_blobs = isinstance(blob_set, list) and len(
            blob_set) > 0 and isinstance(blob_set[0], list)

        # verify layout if a complex set
        if per_set_blobs:
            first_element_blobs = blob_set[0]
            first_query_blobs = first_element_blobs[0]
            # If someone is looking for info logging from PQS, it is likely that blobs are not being set properly.
            #  The wrapping of blobs in general can be confusing. Best suggestion is looking at a loader.
            logger.info("Blobs for first set = " +
                        str(remove_blobs(blob_set[0])))
            logger.info("First Blob for first set = " +
                        str(remove_blobs(blob_set[0][0])))
            if not isinstance(first_query_blobs, list):
                logger.error(
                    "Expected a list of lists for the first element's blob sets")
                expected_list = remove_blobs(first_query_blobs)
                logger.error(
                    f"Blob input for set is : {expected_list}, but should be a list")
                raise Exception(
                    "Could not determine blob strategy; first batch element doesn't have a list of blobs for the first query set. Are you missing a list wrapping in the CVS parser?")

            for set_i in range(len(first_element_blobs)):
                expected_set = first_element_blobs[set_i]
                if not isinstance(expected_set, list):
                    without_blobs = remove_blobs(expected_set)
                    logger.error(
                        f"Blob input for set {set_i} for first item is : {without_blobs}, but should be a list")
                    raise Exception(
                        f"Could not determine blob strategy; first batch element doesn't have a list of blobs for set {set_i}")
                if len(expected_set) != blobs_per_query[set_i]:
                    without_blobs = remove_blobs(expected_set)
                    logger.error(
                        f"Blob input for set {set_i} for first item is : {without_blobs}, but expecting a length of {blobs_per_query[set_i]}")
                    raise Exception(
                        f"Could not executed blob strategy; first batch element's list length doesn't match blob_per_query for the set {set_i} ( {len(expected_set)} != {blobs_per_query[set_i]} )")

        # define method for extracting the blobs
        def blob_filter(all_blobs, strike_list, set_nm): return all_blobs

        # set blob filter based on how blobs are passed to function
        if per_set_blobs:
            # if each query_set has blobs, we must extract out the n-th element in each blob set

            def set_blob_filter(all_blobs, strike_list, set_nm):
                # the list comprehension pulls out the blob set for the requested set
                # the blob set is then flattened as the query expects a flat array using blobs_per_query as the iterator
                # the flat list is them zipped with the strike list, which determines which blobs are unused
                # the filter checks if the blob is to be struc
                # the map pulls the remaining blobs out

                return list(map(lambda pair: pair[0],
                            filter(lambda pair: pair[1] is not True,
                                   itertools.zip_longest(
                                itertools.chain(*[blob_set[set_nm]
                                                for blob_set in all_blobs]),
                                strike_list)
                )
                ))
            blob_filter = set_blob_filter
        else:
            def first_only_blobs(all_blobs, strike_list, set_nm):
                if set_nm == 0:
                    # same as above, without the list comprehension and flattening.
                    return list(map(lambda pair: pair[0], filter(lambda pair: pair[1] is not False,   itertools.zip_longest(all_blobs, strike_list))))
                else:
                    return []
            blob_filter = first_only_blobs

        # start execution
        stored_results = {}
        db_results = []
        db_blobs = None
        for i in range(0, set_total):

            # total number of blobs in this set
            # passing in [] allows for no strikes, so maximum size.
            blobs_this_set = len(blob_filter(blob_set, [], i))
            expected_blobs = blobs_per_query[i] * batch_size
            logger.info(
                f"Set {i}: Commands per query = {commands_per_query[i]}, Blobs per query = {blobs_per_query[i]}"
            )
            if blobs_this_set != expected_blobs:
                logger.error(
                    f"Set {i}: Expected {expected_blobs} blobs, but filter is returning {blobs_this_set}"
                )

            # now we determine if the executing set has a constraint
            # in the head

            # allowed layouts for commands other than the seed command
            # { "cmd" : {} } -> standard single command
            # [{ "cmd1": {}, "cmd2} : {}] -> standard multiple command
            # [{ "constraint" : {} , { "cmd" : {} }] -> constraint with a single command
            # [{ "constraints: {} , [{"cmd1" : {} }, {"cmd2": {} }]] -> constraint with multiple command

            known_constraint_keys = ["results", "apply"]
            constraints = None
            # define filter to optionally remove constraints when passing to executor
            def query_filter(entity, entity_results): return entity[i]

            if i != 0 and isinstance(first_element[i], list) and len(first_element[i]) > 0:
                set_component = first_element[i][0]
                if not isinstance(set_component, list) and any([k in known_constraint_keys for k in set_component.keys()]):
                    constraints = set_component

            if constraints is not None:
                # operations for constraints
                def check_apply_constraint(test_op, item_a, item_b):
                    if DEBUG_CONSTRAINTS:
                        logger.debug(
                            f"check_apply_constraint {test_op} {item_a} {item_b}")
                    if test_op == "==":
                        return item_a == item_b
                    elif test_op == ">":
                        return item_a > item_b
                    elif test_op == "<":
                        return item_a < item_b
                    elif test_op == "!=":
                        return item_a != item_b
                    else:
                        raise Exception(
                            f"Unhandled constraint {test_op} in check_apply_constraint")

                known_constraint_keywords = ['results', 'apply']
                # function called for each row in the set to decide if a query is executed

                def constraint_filter(single_line, single_results):
                    current_constraints = single_line[i][0]

                    if DEBUG_CONSTRAINTS:
                        logger.debug(f"constraint = {current_constraints}")
                        logger.debug(f"results = {single_results}")

                    if 'results' in current_constraints:
                        result_constraints = current_constraints['results']
                        passed_all_constraints = True
                        for result_number in result_constraints:

                            if len(single_results) < result_number or single_results[result_number] is None:
                                # in theory here we have two possibilities: a user can have a correctly formed constraint which didn't execute by design
                                # ( which is what process here )
                                #  or they can have a mis-configured constraint, which is harder to detect.
                                if DEBUG_CONSTRAINTS:
                                    logger.debug(
                                        f"Failed applying constraints, requested results from operation {result_number}, but none existed ( Query didn't run )")
                                passed_all_constraints = False
                                break

                            # query this constraint references
                            prev_query = single_line[result_number]

                            if DEBUG_CONSTRAINTS:
                                logger.debug(
                                    f"prev_query = {single_line[result_number]}")

                            if isinstance(prev_query, dict):
                                prev_query_cmd = [
                                    cmd for cmd in prev_query.keys()][0]
                            elif isinstance(prev_query, list) and isinstance(prev_query[0], dict) \
                                    and all(map(lambda k: k in known_constraint_keywords, prev_query[0].keys()))  \
                                    and isinstance(prev_query[1], dict):
                                prev_query_cmd = [
                                    cmd for cmd in prev_query[1].keys()][0]
                            else:
                                raise Exception(
                                    f"Contraints only implemented with with single queries; query {result_number} not single item.")

                            target_results = single_results[result_number][prev_query_cmd]
                            target_constraints = result_constraints[result_number]
                            for result_item in target_constraints:
                                test = target_constraints[result_item]
                                if not result_item in target_results:
                                    logger.debug(
                                        f"failed results = {target_results}")
                                    raise Exception(
                                        f"Failed applying constraints, requested result '{result_item}' from operation {result_number}, but none exited")
                                # if constraint passes, apply:
                                if not check_apply_constraint(test[0], target_results[result_item], test[1]):
                                    passed_all_constraints = False
                                    break

                        if passed_all_constraints:
                            return single_line[i][1]
                        else:
                            return None
                    elif 'apply' in current_constraints:
                        # apply means run the line
                        return single_line[i][1]
                    else:
                        raise Exception(
                            f"incorrectly formatted constraint; no known conatraint action")

                query_filter = constraint_filter

            local_success_statuses = [0, 2]

            # queries are by row first, so we run query_filter on each query
            # we pass the entire row's data, then we retrieve all of the stored results for that row

            queries = [query_filter(query_set[row_num], [stored_results[res_grp][row_num] for res_grp in stored_results])
                       for row_num in range(0, len(query_set))]

            # finally, we remove queries that were reduced to None so the base executor doesn't have to deal with them
            # and unwrap multiple commands per element into a flat list; map allows one command to process either type
            #  by wrapping all in a list.
            executable_queries = list(itertools.chain(*
                                                      map(lambda q: q if isinstance(q, list) else [q],
                                                          filter(lambda q: q is not None, queries))
                                                      ))

            # produce strike list based on queries that are none
            blob_strike_list = list(map(lambda q: q is None, queries))

            # filter out struck blobs
            used_blobs = filter(lambda b: b is not None,
                                blob_filter(blob_set, blob_strike_list, i))

            # TODO: add wrapped response_handler.
            if response_handler != None:
                logger.warning(
                    "ParallelQuerySet does not yet support a response_handler which will identify which set is being worked on")
            if len(executable_queries) > 0:
                result_code, db_results, db_blobs = base_executor(executable_queries, used_blobs,
                                                                  db, local_success_statuses,
                                                                  None, commands_per_query[i], blobs_per_query[i], strict_response_validation=strict_response_validation)
            else:
                logger.info(
                    f"Skipped executing set {i}, no executable queries")

            if result_code == 1:
                logger.error(
                    f"Ran into error on set {i} in ParallQuerySet, unable to continue")
                return 1, db_results, db_blobs

            # expand results so queries that didn't run show up as a None
            results_off = 0

            def insert_empty_results(result_value):
                nonlocal results_off
                if result_value is None:
                    return None
                else:
                    results_off += 1
                    return db_results[(results_off - 1)]

            stored_results[i] = [insert_empty_results(q) for q in queries]

        return 0, db_results, db_blobs  # end execute_batch_sets
    return execute_batch_sets


class ParallelQuerySet(ParallelQuery):
    """
    **Parallel and Batch Set Multi-Querier for ApertureDB**
    This class provides the mechanism to run multiple queries over a single csv.
    Per-query actions are done by ParallelQuery.

    Args:
        db (Connector): The ApertureDB Connector
        dry_run (bool, optional): If True, no queries are executed. Defaults to False.
    """

    def __init__(self, db: Connector, dry_run: bool = False):

        super().__init__(db, dry_run)

        self.base_batch_command = self.batch_command

    def verify_generator(self, generator) -> bool:
        # first level should be grouping of commands
        # first cmd should have a list of query sets
        if isinstance(generator[0], list) or isinstance(generator[0], tuple):
            cmd = generator[0]
            if isinstance(generator[0][0], list):
                return True

        logger.error(
            f"Could not determine query structure from:\n{generator[0]}")
        logger.error(type(generator[0]))
        return False

    def do_batch(self, db: Connector, data: List[Tuple[Commands, Blobs]]) -> None:
        """
        This is an override of ParallelQuery.do_batch.

        This is the per-worker function which is the entry-point to a unit of work.

        Args:
            db (Connector): The ApertureDB Connector
            data (List[Tuple[Query, Blobs]]): A list of tuples, each containing a list of commands and a list of blobs
        """
        if not hasattr(self.generator, "commands_per_query"):
            raise Exception(
                "Generator must specify commands per query for ParallelQuerySet")
        if not hasattr(self.generator, "blobs_per_query"):
            raise Exception(
                "Generator must specify blobs per query for ParallelQuerySet")
        self.commands_per_query = self.generator.commands_per_query
        self.blobs_per_query = self.generator.blobs_per_query
        set_response_handler = None
        if hasattr(self.generator, "set_response_handler") and callable(self.generator.set_response_handler):
            set_response_handler = self.generator.set_response_handler
        self.batch_command = gen_execute_batch_sets(
            self.base_batch_command, set_response_handler)

        ParallelQuery.do_batch(self, db, data)

    def print_stats(self) -> None:

        times = np.array(self.times_arr)
        total_queries_exec = len(times)

        # Todo - implement per-set data.

        print("============ ApertureDB Parallel Query Set Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

        print("=========================================================")



import time
import requests

import numpy as np

import logging

logger = logging.getLogger(__name__)


class Sources():
    """
    **Load data from various resources**
    """

    def __init__(self, n_download_retries, **kwargs):

        self.n_download_retries = n_download_retries

        # Use custom clients if specified
        self.s3 = None if "s3_client" not in kwargs else kwargs["s3_client"]
        self.http_client = requests.Session(
        ) if "http_client" not in kwargs else kwargs["http_client"]

    def load_from_file(self, filename):
        """
        Load data from a file.
        """
        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except Exception as e:
            logger.error(f"VALIDATION ERROR: {filename}")
            logger.exception(e)
        finally:
            if not fd.closed:
                fd.close()
        return False, None

    def load_from_http_url(self, url, validator):
        """
        Load data from a http url.
        """
        retries = 0
        while True:
            imgdata = self.http_client.get(url)
            if imgdata.ok:
                imgbuffer = np.frombuffer(imgdata.content, dtype='uint8')
                if not validator(imgbuffer):
                    logger.error(f"VALIDATION ERROR: {url}")
                    return False, None

                return imgdata.ok, imgdata.content
            else:
                if retries >= self.n_download_retries:
                    break
                logger.warning(f"Retrying object: {url}")
                retries += 1
                time.sleep(2)

        return False, None

    def load_from_s3_url(self, s3_url, validator):
        retries = 0

        while True:
            try:
                bucket_name = s3_url.split("/")[2]
                object_name = s3_url.split("s3://" + bucket_name + "/")[-1]
                s3_response_object = self.s3.get_object(
                    Bucket=bucket_name, Key=object_name)
                img = s3_response_object['Body'].read()
                imgbuffer = np.frombuffer(img, dtype='uint8')
                if not validator(imgbuffer):
                    logger.error(f"VALIDATION ERROR: {s3_url}")
                    return False, None

                return True, img
            except Exception as e:
                if retries >= self.n_download_retries:
                    break
                logger.warning(f"Retrying object: {s3_url}", exc_info=True)
                retries += 1
                time.sleep(2)

        logger.error(f"S3 ERROR: {s3_url}")
        return False, None

    def load_from_gs_url(self, gs_url, validator):
        retries = 0
        from google.cloud import storage
        client = storage.Client()
        while True:
            try:
                bucket_name = gs_url.split("/")[2]
                object_name = gs_url.split("gs://" + bucket_name + "/")[-1]

                blob = client.bucket(bucket_name).blob(
                    object_name).download_as_bytes()
                imgbuffer = np.frombuffer(blob, dtype='uint8')
                if not validator(imgbuffer):
                    logger.warning(f"VALIDATION ERROR: {gs_url}")
                    return False, None
                return True, blob
            except:
                if retries >= self.n_download_retries:
                    break
                logger.warning("Retrying object: {gs_url}", exc_info=True)
                retries += 1
                time.sleep(2)

        logger.error(f"GS ERROR: {gs_url}")
        return False, None



from __future__ import annotations

from enum import Enum
from typing import List, Tuple

from aperturedb.Constraints import Constraints
from aperturedb.Operations import Operations
from aperturedb.Sort import Sort
from pydantic import BaseModel


class ObjectType(Enum):
    """
    Based on : [API description](/query_language/Overview/API%20Description)

    """
    BLOB = "_Blob"
    BOUNDING_BOX = "_BoundingBox"
    CONNECTION = "_Connection"
    DESCRIPTOR = "_Descriptor"
    DESCRIPTORSET = "_DescriptorSet"
    ENTITY = "_Entity"
    FRAME = "_Frame"
    IMAGE = "_Image"
    POLYGON = "_Polygon"
    VIDEO = "_Video"


class Config(Enum):
    SAVE_NAME = 1
    SAVE_VALUE = 2


config = Config.SAVE_NAME


def generate_save_query(
        obj: BaseModel,
        cached: List[str] = None,
        source_field: str = None,
        index: int = 1,
        parent: int = 0) -> Tuple(List[object], List[bytearray], str):
    """
    Takes the user model, and builds out a sequence of commands that creates
    a similar structure on apertureDB's graph.

    Args:
        obj (BaseModel): The object from the user domain.
        cached (List[str], optional): helps to optimize sending one blob per node. Defaults to None.
        source_field (str, optional): Preserves the relevant connection information from user objects. Defaults to None.
        index (int, optional): The index to start creating references from. Defaults to 1.
        parent (int, optional): The parent of the current node. Defaults to None.

    Returns:
        tuple : Contains 3 items:
            * List of commands with current node and its children
            * List of blobs with current node and its children
            * index, adjusting the count of children + self for the number of commands
              in current query
    """
    if cached == None:
        cached = []
    query = []
    dependents = []
    props = {}
    blobs = []
    cindex = index
    if obj.id not in cached:
        for p in obj.__dict__.keys():
            if "_" == p[0]:
                continue
            subobj = getattr(obj, p)
            if isinstance(subobj, int) or isinstance(subobj, str):
                props[p] = subobj
            elif isinstance(subobj, Enum):
                props[p] = subobj.name if config == Config.SAVE_NAME else subobj.value
            elif isinstance(subobj, list):
                for i, si in enumerate(subobj):
                    q, b, index = generate_save_query(
                        si, cached=cached, source_field=f"{p}[{i}]", index=index + 1, parent=cindex)
                    blobs.extend(b)
                    dependents.extend(q)
            else:
                q, b, index = generate_save_query(
                    subobj, cached=cached, source_field=f"{p}", index=index + 1, parent=cindex)
                dependents.extend(q)
                blobs.extend(b)

    # We still want to create dummy Add* commands so that references
    # for creating already created instances are mantained.
    params = {
        "_ref": cindex,
        "properties": props if "id" in props else {},
        "if_not_found": {
            "id": ["==", props["id"] if "id" in props else obj.id]
        }
    }
    if hasattr(obj, "type"):
        props.pop('type', None)
        if obj.type != ObjectType.ENTITY:
            props["user_type"] = type(obj).__name__
            # Generates a Add<ADB object> command
            if obj.id not in cached:
                query.append(
                    QueryBuilder.add_command(obj.type.value, params=params))
            else:
                params["constraints"] = params["if_not_found"]
                params.pop("if_not_found", None)
                params.pop("properties", None)
                query.append(
                    QueryBuilder.find_command(obj.type.value, params=params))
        if obj.type in [ObjectType.IMAGE, ObjectType.VIDEO, ObjectType.BLOB]:
            # Do not send blob, if Node has been added to set of commands.
            if obj.id not in cached:
                with open(obj.file, "rb") as instream:
                    blobs.append(instream.read())
                props.pop('file', None)
        else:
            # Generates an AddEntity
            query.append(
                QueryBuilder.add_command(type(obj).__name__, params=params))

        cached.append(obj.id)
    if parent > 0:
        params = {
            "src": parent,
            "dst": cindex,
            "class": type(obj).__name__,
            "properties": {
                "source_field": source_field
            }
        }
        query.append(
            QueryBuilder.add_command(
                ObjectType.CONNECTION.value, params=params)
        )
    query.extend(dependents)
    return query, blobs, index


class QueryBuilder():
    @classmethod
    def find_command(self, oclass: str, params: dict) -> dict:
        return self.build_command(oclass, params, "Find")

    @classmethod
    def add_command(self, oclass: str, params: dict) -> dict:
        return self.build_command(oclass, params, "Add")

    @classmethod
    def build_command(self, oclass, params, operation):
        command = {
            f"{operation}Entity": params
        }
        members = [m.value for m in ObjectType]
        if oclass.startswith("_"):
            if oclass in members:
                command = {
                    f"{operation}{oclass[1:]}": params
                }
            else:
                raise Exception(
                    f"Invalid Object type. Should not begin with _, except for {members}")
        else:
            if operation == "Find":
                params["with_class"] = oclass
            else:
                params["class"] = oclass
        return command


class Query():
    """
    This is the underlying class to generate a query using python code.
    """
    db_object = "Entity"
    next = None

    def connected_to(self,
                     spec: Query,
                     adj_to: int = 0) -> Query:
        """
        Can be used to connect two commands together, for non trivial queries.
        This is the lower level function to create a query.

        Args:
            spec (Query): One command of the query.
            adj_to (int, optional): The command to refer to. Defaults to 0.

        Returns:
            Query: The completed, connected query.
        """
        spec.adj_to = self.adj_to + 1 if adj_to == 0 else adj_to
        self.next = spec
        return self

    def command_properties(self, prop: str = "") -> List[str]:
        """
        Helper function to get the properties of all commands the query.

        Args:
            v (str, optional): _description_. Defaults to "".

        Returns:
            List[str]: Properties from commands in the order they should be executed.
        """
        chain = []
        p = self
        while p is not None:
            chain.append(getattr(p, prop))
            p = p.next
        return chain

    @classmethod
    def spec(cls,
             constraints: Constraints = None,
             operations: Operations = None,
             with_class: str = "",
             limit: int = -1,
             sort: Sort = None,
             list: List[str] = None,
             group_by_src: bool = False,
             blobs: bool = False
             ) -> Query:
        """
        The [specification](/query_language/Overview/API%20Description) for a command to be used in a query.

        Args:
            constraints (Constraints, optional): [Constraints](/query_language/Reference/shared_command_parameters/constraints) . Defaults to None.
            with_class (ObjectType, optional): _description_. Defaults to ObjectType.CUSTOM.
            limit (int, optional): _description_. Defaults to -1.
            sort (Sort, optional): _description_. Defaults to None.
            list (List[str], optional): _description_. Defaults to None.

        Returns:
            Query: The query object.
        """
        return Query(
            constraints=constraints,
            operations=operations,
            with_class=with_class,
            limit=limit,
            sort = sort,
            list = list,
            blobs=blobs,
            group_by_src = group_by_src
        )

    def __init__(self,
                 constraints: Constraints = None,
                 operations: Operations = None,
                 with_class: str = "",
                 limit: int = -1,
                 sort: Sort = None,
                 list: List[str] = None,
                 group_by_src: bool = False,
                 blobs: bool = False,
                 adj_to: int = 0):
        self.constraints = constraints
        self.operations = operations
        self.with_class = with_class
        self.limit = limit
        self.sort = sort
        self.list = list
        self.group_by_src = group_by_src
        self.blobs = blobs
        self.adj_to = adj_to + 1

    def query(self) -> List[dict]:
        results_section = "results"
        cmd_params = {results_section: {}}
        if self.limit != -1:
            cmd_params[results_section]["limit"] = self.limit
        if self.sort:
            cmd_params[results_section]["sort"] = self.sort._sort
        if self.list is not None and len(self.list) > 0:
            cmd_params[results_section]["list"] = self.list
        else:
            cmd_params[results_section]["all_properties"] = True
        cmd_params[results_section]["group_by_source"] = self.group_by_src

        if self.constraints:
            cmd_params["constraints"] = self.constraints.constraints
        if self.operations:
            cmd_params["operations"] = self.operations.operations_arr

        self.with_class = self.with_class if self.db_object == "Entity" else self.db_object
        cmd = QueryBuilder.find_command(
            oclass=self.with_class, params=cmd_params)
        self.find_command = list(cmd.keys())[0]
        query = [cmd]
        if self.next:
            next_commands = self.next.query()
            list(next_commands[0].values())[0]["is_connected_to"] = {
                "ref": self.adj_to
            }
            query.extend(next_commands)
        return query



from __future__ import annotations
from pydantic import BaseModel, Field
from typing_extensions import Annotated
from typing import ClassVar
from uuid import uuid4
from aperturedb.Query import ObjectType


class IdentityModel(BaseModel):
    id: Annotated[str, Field(default_factory=lambda: uuid4().hex)]
    # Change as per the docs for the error
    # https://docs.pydantic.dev/dev-v2/usage/errors/#model-field-overridden
    type: ClassVar[ObjectType] = ObjectType.ENTITY


class BlobModel(IdentityModel):
    file: Annotated[str, Field(title="Filepath", description="Path to file")]
    type = ObjectType.BLOB


class ImageModel(BlobModel):
    type = ObjectType.IMAGE


class VideoModel(BlobModel):
    type = ObjectType.VIDEO


class DesctiptorModel(BlobModel):
    type = ObjectType.DESCRIPTOR


class PolygonModel(IdentityModel):
    type = ObjectType.POLYGON


class FrameModel(IdentityModel):
    type = ObjectType.FRAME


class DescriptorSetModel(IdentityModel):
    type = ObjectType.DESCRIPTORSET


class BoundingBoxModel(IdentityModel):
    type = ObjectType.BOUNDING_BOX




from typing import List, Tuple
import os
import pandas as pd
from kaggle.api.kaggle_api_extended import KaggleApi
import zipfile
from aperturedb.Subscriptable import Subscriptable


class KaggleData(Subscriptable):
    """
    **Class to wrap around a Dataset retrieved from kaggle**

    A DataSet downloaded from kaggle does not implement a standard mechanism to iterate over its values
    This class intends to provide an abstraction like that of a pytorch dataset
    where the iteration over Dataset elements yields an atomic record.

    :::note
    This class should be subclassed with specific implementations of generate_index and generate_query.
    :::

    Example subclass: [CelebADataKaggle](https://github.com/aperture-data/aperturedb-python/blob/develop/examples/CelebADataKaggle.py)

    Args:
        dataset_ref (str): URL of kaggle dataset, for example https://www.kaggle.com/datasets/jessicali9530/celeba-dataset
        records_count (int): number of records to provide to generate.

    """

    def __init__(
            self,
            dataset_ref: str,
            records_count: int = -1) -> None:
        self._collection = None
        self.records_count = records_count
        kaggle = KaggleApi()
        kaggle.authenticate()
        if "datasets/" in dataset_ref:
            dataset_ref = dataset_ref[dataset_ref.index(
                "datasets/") + len("datasets/"):]

        workdir = os.path.join("kaggleds", dataset_ref)

        files = kaggle.dataset_list_files(dataset_ref)

        # do not unzip from kaggle's API as it deletes the archive and
        # a subsequent run results in a redownload.
        x = kaggle.dataset_download_files(
            dataset=dataset_ref,
            path=workdir,
            quiet=False,
            unzip=False)

        archive = None
        for _, subdirs, dfiles in os.walk(workdir):
            if len(dfiles) == 1 and len(subdirs) == 0:
                archive = os.path.join(workdir, dfiles[0])

                with zipfile.ZipFile(archive, 'r') as zip_ref:
                    zip_ref.extractall(workdir)

                break
        self.workdir = workdir
        self.collection = self.generate_index(
            workdir, self.records_count).to_dict('records')

    def getitem(self, subscript):
        return self.generate_query(subscript)

    def __len__(self):
        return len(self.collection)

    def generate_index(self, root: str, records_count: int = -1) -> pd.DataFrame:
        """**Generate a way to access each record downloaded at the root**

        Args:
            root (str): Path to wich kaggle downloads a Dataset.

        Returns:
            pd.DataFrame: The Data loaded in a dataframe.
        """
        raise Exception("To be implemented by subclass")

    def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
        """
        **Takes information from one atomic record from the Data and converts it to Query for apertureDB**

        Args:
            idx (int): index of the record in collection.

        Raises:
            Exception: _description_

        Returns:
            Tuple[List[dict], List[bytes]]: A pair of list of commands and optional list of blobs to go with them.
        """
        raise Exception("To be implemented by subclass")



from __future__ import annotations
from aperturedb.Entities import Entities
from aperturedb.ParallelQuery import execute_batch


class Polygons(Entities):
    db_object = "_Polygon"

    def intersection(self, other: Polygons, threshold: float) -> Polygons:
        """
        Find a set of polygons that intersect with another set of polygons.
        The threshold is user specified and is used to determine if two polygons
        sufficiently overlap to be considered intersecting.

        Args:
            other (Polygons): Set of polygons to intersect with.
            threshold (float): The threshold for determining if two polygons are sufficiently intersecting.

        Returns:
            Polygons: unique set of polygons that intersect with the other set of polygons.
        """
        result = set()
        for p1 in self:
            for p2 in other:
                query = [
                    {
                        "FindEntity": {
                            "_ref": 1,
                            "unique": True,
                            "constraints": {
                                "_uniqueid": ["==", p1["_uniqueid"]]
                            }
                        }
                    }, {
                        "FindEntity": {
                            "_ref": 2,
                            "unique": True,
                            "constraints": {
                                "_uniqueid": ["==", p2["_uniqueid"]]
                            }
                        }
                    }, {
                        "RegionIoU": {
                            "roi_1": 1,
                            "roi_2": 2,
                        }
                    }
                ]
                res, r, b = execute_batch(query, [], self.db)
                if r[2]["RegionIoU"]["IoU"][0][0] > threshold:
                    result.add(int(p1["ann_id"]))
                    result.add(int(p2["ann_id"]))
        return list(result)



import math
import time
from threading import Thread


from aperturedb import ProgressBar


class Parallelizer:
    """**Generic Parallelizer**

    A parallelizer converts a series of operations to be executed and partitions it into
    batches, to be executed by multiple threads of execution.
    ```mermaid
    gantt
        title Parallel execution
        dateFormat HH:mm:ss
        section Worker1
        Batch1 :w1, 00:00:00, 10s
        Batch3 :w3, after w1, 10s
        Batch5 :after w3, 10s

        section Worker2
        Batch2 :w2, 00:00:00, 10s
        Batch4 :w4, after w2, 10s
        Batch6 :w6, after w4, 10s

    ```
    """

    def __init__(self, progress_to_file=""):

        self.pb_file = progress_to_file

        self._reset()

    def _reset(self, batchsize: int = 1, numthreads: int = 1):

        # Default Values
        self.batchsize = batchsize
        self.numthreads = numthreads

        self.total_actions = 0
        self.times_arr = []
        self.total_actions_time = 0
        self.error_counter = 0
        self.actual_stats = []

        if self.pb_file:
            self.pb = ProgressBar.ProgressBar(self.pb_file)
        else:
            self.pb = ProgressBar.ProgressBar()

    def get_times(self):

        return self.times_arr

    def run(self, generator, batchsize: int, numthreads: int, stats: bool):

        self._reset(batchsize, numthreads)
        self.stats = stats
        self.generator = generator
        if hasattr(generator, "sample_count"):
            print("sample_count", generator.sample_count)
            self.total_actions = generator.sample_count
        else:
            self.total_actions = len(generator)

        start_time = time.time()

        if self.total_actions < batchsize:
            elements_per_thread = self.total_actions
            self.numthreads = 1
        else:
            elements_per_thread = math.ceil(
                self.total_actions / self.numthreads)

        thread_arr = []
        for i in range(self.numthreads):
            idx_start = i * elements_per_thread
            idx_end = min(idx_start + elements_per_thread,
                          self.total_actions)

            thread_add = Thread(target=self.worker,
                                args=(i, generator, idx_start, idx_end))
            thread_arr.append(thread_add)

        a = [th.start() for th in thread_arr]
        a = [th.join() for th in thread_arr]

        # Update progress bar to completion
        if self.stats:
            self.pb.update(1)

        self.total_actions_time = time.time() - start_time

        if self.stats:
            self.print_stats()

    def print_stats(self):
        """
            Must be implemented by child class
        """
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: queryMessage.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='queryMessage.proto',
  package='VDMS.protobufs',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x12queryMessage.proto\x12\x0eVDMS.protobufs\":\n\x0cqueryMessage\x12\x0c\n\x04json\x18\x01 \x01(\t\x12\r\n\x05\x62lobs\x18\x02 \x03(\x0c\x12\r\n\x05token\x18\x03 \x01(\tb\x06proto3'
)




_QUERYMESSAGE = _descriptor.Descriptor(
  name='queryMessage',
  full_name='VDMS.protobufs.queryMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='json', full_name='VDMS.protobufs.queryMessage.json', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='VDMS.protobufs.queryMessage.blobs', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='token', full_name='VDMS.protobufs.queryMessage.token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=96,
)

DESCRIPTOR.message_types_by_name['queryMessage'] = _QUERYMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

queryMessage = _reflection.GeneratedProtocolMessageType('queryMessage', (_message.Message,), {
  'DESCRIPTOR' : _QUERYMESSAGE,
  '__module__' : 'queryMessage_pb2'
  # @@protoc_insertion_point(class_scope:VDMS.protobufs.queryMessage)
  })
_sym_db.RegisterMessage(queryMessage)


# @@protoc_insertion_point(module_scope)



import logging
from aperturedb.Blobs import Blobs
from aperturedb.BoundingBoxes import BoundingBoxes
from aperturedb.Constraints import Constraints
from aperturedb.Entities import Entities
from aperturedb.Images import Images
from aperturedb.Query import ObjectType, Query
from aperturedb.Query import generate_save_query
from aperturedb.QueryTypes import ImageModel, IdentityModel
import random
from typing import List
from enum import Enum
from devtools import debug

logger = logging.getLogger(__file__)


class TestDataWizardEntities():
    def test_get_persons(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(in_csv_file="./input/persons.adb.csv")
        all_persons = Entities.retrieve(db,
                                        spec=Query.spec(with_class="Person"))
        assert len(all_persons) == len(loaded)
        too_young = all_persons.filter(lambda p: p["age"] < 18)
        too_old = all_persons.filter(lambda p: p["age"] > 60)
        working_people = all_persons - too_old - too_young

        assert all([p["age"] >= 18 and p["age"] <= 60 for p in working_people])
        ordered = working_people.sort(key=lambda p: p["age"])
        youngest_working_five = ordered[:5]
        eldest_working_five = ordered[-5:]

        assert len(youngest_working_five) == 5
        assert len(eldest_working_five) == 5
        assert youngest_working_five[0]["age"] < eldest_working_five[0]["age"]

    def test_get_persons_constraints(self, retired_persons):
        logger.info(f"Retired persons count = {len(retired_persons)}")
        assert all(p["age"] >= 60 for p in retired_persons)

    def test_get_persons_display(self, retired_persons):
        logger.info(f"Retired persons count = {len(retired_persons)}")
        df = retired_persons.inspect()
        logger.info(f"\n{df}")
        ages = df['age']
        assert ages[ages >= 60].count() == len(df)

    def test_update_properties(self, db, retired_persons):
        risk_factors = [{
            "risk_factor": (p["age"] - 60) // 10
        } for p in retired_persons]
        retired_persons.update_properties(risk_factors)
        df = retired_persons.inspect()
        logger.info(f"\n{df}")
        persons = Entities.retrieve(db,
                                    spec=Query.spec(with_class="Person",
                                                    constraints=Constraints().greaterequal("age", 60)))

        assert len(persons) == len(retired_persons)
        assert all(p["risk_factor"] == (p["age"] - 60) // 10 for p in persons)


class TestDataWizardImages():
    def test_get_images(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(
            "./input/images.adb.csv", rec_count=10)
        images = Images.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            images, Images), f"{type(images)} is not an instance of Images"
        logger.info(f"\n{images.inspect()}")
        assert images != None
        assert len(images) <= 10


class TestDataWizardBlobs():
    def test_get_blobs(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv("./input/blobs.adb.csv", rec_count=10)
        blobs = Blobs.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            blobs, Blobs), f"blobs is not an instance of Blobs = {blobs}"
        logger.info(f"\n{blobs.inspect()}")
        assert blobs != None
        assert len(blobs) <= 10


class TestDataWizardBoundingBoxes():
    def test_get_bounding_boxes(self, insert_data_from_csv, db):
        loaded, _ = insert_data_from_csv(
            "./input/images.adb.csv", rec_count=10)
        loaded, _ = insert_data_from_csv(
            "./input/bboxes.adb.csv", rec_count=10)
        bboxes = BoundingBoxes.retrieve(db, Query.spec(limit=10))
        assert isinstance(
            bboxes, BoundingBoxes), f"bboxes is not an instance of BoundingBoxes = {bboxes}"
        logger.info(f"\n{bboxes.inspect()}")
        assert bboxes != None
        assert len(bboxes) <= 10


def make_people(count: int = 1) -> List[object]:
    class Side(Enum):
        RIGHT = 1
        LEFT = 2

    class Finger(IdentityModel):
        nail_clean: bool = False

    class Hand(ImageModel):
        side: Side = None
        thumb: Finger = None
        fingers: List[Finger] = []

    class Person(IdentityModel):
        name: str = ""
        hands: List[Hand] = []
        dominant_hand: Hand = None

    def make_hand(side: Side) -> Hand:
        hand = Hand(side = side, file= "input/images/0079.jpg")
        hand.fingers = [Finger(nail_clean=True) if random.randint(
            0, 1) == 1 else Finger(nail_clean=False) for i in range(5)]
        hand.thumb = hand.fingers[0]
        return hand

    people = []
    for i in range(10):
        person = Person(name=f"adam{i+1}")
        left_hand = make_hand(Side.LEFT)
        right_hand = make_hand(Side.RIGHT)
        person.hands.extend([left_hand, right_hand])
        person.dominant_hand = person.hands[0]
        people.append(person)
    return people


class TestQueryBuilder():
    def test_all_info_preserved(self):
        # this will create 10 people entities.
        # Each person will have 2 hands (+1 connection via dominant hand)
        # Each hand will have 10 fingers (+2 connections via thumb)
        # Everything will be interconnected.
        # Nodes per person = 1 + 3 + 12 = 16 (The commands will be generated with if not found)
        # Connections per person = 3 for hand 12 for fingers
        # Commands = 16 + 15 = 31
        people = make_people(10)
        total_commands = []
        total_blobs = []
        for person in people:
            q, b, current_ref = generate_save_query(person)
            total_commands += q
            total_blobs += b
            # 16 entities have been inserted and referenced
            assert current_ref == 16

        assert len(total_commands) == 310
        assert len(total_blobs) == 20
        assert len(
            list(filter(lambda cmd: "AddConnection" in cmd, total_commands))) == 150
        assert len(
            list(filter(lambda cmd: "AddEntity" in cmd, total_commands))) == 130
        assert len(
            list(filter(lambda cmd: "AddImage" in cmd, total_commands))) == 20
        assert len(
            list(filter(lambda cmd: "FindImage" in cmd, total_commands))) == 10



from matplotlib.transforms import Bbox
from aperturedb import ParallelLoader
from aperturedb import CSVParser

HEADER_X_POS = "x_pos"
HEADER_Y_POS = "y_pos"
HEADER_WIDTH = "width"
HEADER_HEIGHT = "height"
IMG_KEY_PROP = "img_key_prop"
IMG_KEY_VAL = "img_key_value"


class BBoxDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB BBox Data.**

    This class loads the Bounding Box Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.

    :::note Is backed by a CSV file with the following columns:
    ``IMG_KEY``, ``x_pos``, ``y_pos``, ``width``, ``height``, ``BBOX_PROP_NAME_1``, ... ``BBOX_PROP_NAME_N``, ``constraint_BBOX_PROP_NAME_1``
    :::

    **IMG_KEY**: column has the property name of the image property that
    the bounding box will be connected to, and each row has the value
    that will be used for finding the image.

    **x_pos, y_pos**: Specify the coordinates of top left of the bounding box.

    **width, height**: Specify the dimensions of the bounding box, as integers (unit is in pixels).

    **BBOX_PROP_NAME_N**: is an arbitrary name of the property of the bounding
    box, and each row has the value for that property.

    **constraint_BBOX_PROP_NAME_1**: Constraints against specific property, used for conditionally adding a Bounding Box.

    Example CSV file::

        img_unique_id,x_pos,y_pos,width,height,type,dataset_id,constraint_dataset_id
        d5b25253-9c1e,257,154,84,125,manual,12345,12345
        d5b25253-9c1e,7,537,522,282,manual,12346,12346
        ...

    Example usage:

    ``` python

        data = BBoxDataCSV("/path/to/BoundingBoxesData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```

    :::info
    In the above example, the constraint_dataset_id ensures that a bounding box with the specified
    dataset_id would be only inserted if it does not already exist in the database.
    :::

    """

    def __init__(self, filename: str, **kwargs):

        super().__init__(filename, **kwargs)

        self.props_keys = [x for x in self.header[5:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[5:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]

        self.img_key = self.header[0]
        self.command = "AddBoundingBox"

    def get_indices(self):
        return {
            "entity": {
                "_BoundingBox": self.get_indexed_properties()
            }
        }

    def getitem(self, idx):
        q = []
        img_id = self.df.loc[idx, self.img_key]
        fi = {
            "FindImage": {
                "_ref": 1,
                "unique": True,
                "constraints": {
                    self.img_key: ["==", img_id],
                },
                "blobs": False,
            },
        }
        q.append(fi)

        box_data_headers = [HEADER_X_POS,
                            HEADER_Y_POS, HEADER_WIDTH, HEADER_HEIGHT]
        box_data = [int(self.df.loc[idx, h]) for h in box_data_headers]

        rect_attrs = ["x", "y", "width", "height"]
        custom_fields = {
            "image_ref": 1,
            "rectangle": {
                attr: val for attr, val in zip(rect_attrs, box_data)
            },
        }
        abb = self._basic_command(idx, custom_fields)

        properties = self.parse_properties(idx)
        if properties:
            props = properties
            if "_label" in props:
                abb[self.command]["label"] = props["_label"]
                props.pop("_label")
            # Check if props is not empty after removing "_label"
            if props:
                abb[self.command]["properties"] = props
        q.append(abb)

        return q, []

    def validate(self) -> None:

        self.header = list(self.df.columns.values)

        if self.header[1] != HEADER_X_POS:
            raise Exception("Error with CSV file field: " + HEADER_X_POS)
        if self.header[2] != HEADER_Y_POS:
            raise Exception("Error with CSV file field: " + HEADER_Y_POS)
        if self.header[3] != HEADER_WIDTH:
            raise Exception("Error with CSV file field: " + HEADER_WIDTH)
        if self.header[4] != HEADER_HEIGHT:
            raise Exception("Error with CSV file field: " + HEADER_HEIGHT)



from dataclasses import dataclass


@dataclass(repr=False)
class Configuration:
    """
    **Configuration object for ApertureDB sdk to be able to connect to ApertureDB**
    """
    host: str
    port: int
    username: str
    password: str
    name: str
    use_ssl: bool = True
    use_rest: bool = False
    use_keepalive: bool = True
    retry_interval_seconds: int = 1
    # Max number of attempts to retry the initial connection (0 means infinite)
    # This is useful when the aperturedb server is not ready yet.
    retry_max_attempts: int = 3

    def __repr__(self) -> str:
        mode = "REST" if self.use_rest else "TCP"
        return f"[{self.host}:{self.port} as {self.username} using {mode} with SSL={self.use_ssl}]"



import json
import typer
from aperturedb.ParallelQuery import execute_batch
from aperturedb.cli.console import console
from aperturedb.Utils import create_connector

app = typer.Typer()


@app.command()
def from_json_file(filepath: str):
    db = create_connector()
    with open(filepath) as inputstream:
        transaction = json.loads(inputstream.read())
        result, response, blobs = execute_batch(
            db=db,
            q=transaction,
            blobs=[])
        console.log(result)
        console.log(response)



#
# The MIT License
#
# @copyright Copyright (c) 2017 Intel Corporation
# @copyright Copyright (c) 2021 ApertureData Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from __future__ import annotations
from . import queryMessage_pb2
import sys
import os
import socket
import struct
import time
import json
import ssl
import logging
from datetime import datetime, timedelta

import keepalive

from threading import Lock
from types import SimpleNamespace
from dataclasses import dataclass
from aperturedb.Configuration import Configuration

logger = logging.getLogger(__name__)


PROTOCOL_VERSION = 1


class UnauthorizedException(Exception):
    pass


class UnauthenticatedException(Exception):
    pass


@dataclass
class Session():

    session_token:      str
    refresh_token:      str
    session_token_ttl:  int
    refresh_token_ttl:  int
    session_started:    time.time = time.time()

    def valid(self) -> bool:
        session_age = time.time() - self.session_started

        # This triggers refresh if the session is about to expire.
        if session_age > self.session_token_ttl - \
                int(os.getenv("SESSION_EXPIRY_OFFSET_SEC", 10)):
            return False

        return True


class Connector(object):
    """
    **Class to facilitate connections with an instance of ApertureDB**

    It lets the client execute any JSON query based on the [ApertureDB query language specification](/query_language/Overview/API%20Description)

    Args:
        str (host): Address of the host to connect to.
        int (port): Port to connect to.
        str (user): Username to specify while establishing a connection.
        str (password): Password to specify while connecting to the db.
        str (token): Token to use while connecting to the database.
        bool (use_ssl): Use SSL to encrypt communication with the database.
        bool (use_keepalive): Set keepalive on the connection with the database.
            This has two benefits: It reduces the chance of disconnection for a long-running query,
            and it means that disconnections are detected sooner.
            Turn this off to reduce traffic on high-cost network connections.
        Configuration (config): Configuration object to use for connection.
    """

    def __init__(self, host="localhost", port=55555,
                 user="", password="", token="",
                 use_ssl=True,
                 shared_data=None,
                 authenticate=True,
                 use_keepalive=True,
                 retry_interval_seconds=1,
                 retry_max_attempts=3,
                 config: Configuration = None):
        """
        Constructor for the Connector class.
        """
        self.connected = False
        self.last_response = ''
        self.last_query_time = 0
        self.authenticated = False
        self.last_query_timestamp = None
        # suppress connection warnings which occur more than this time
        # after the last query
        self.query_connection_error_suppression_delta = timedelta(seconds=30)

        if config is None:
            self.host = host
            self.port = port
            self.use_ssl = use_ssl
            self.use_keepalive = use_keepalive
            # Create a configuration object, to show better error messages
            self.config = Configuration(
                host=self.host,
                port=self.port,
                use_ssl=self.use_ssl,
                username=user,
                password=password,
                name="runtime",
                use_keepalive=use_keepalive,
                retry_interval_seconds=retry_interval_seconds,
                retry_max_attempts=retry_max_attempts
            )
        else:
            self.config = config
            self.host = config.host
            self.port = config.port
            self.use_ssl = config.use_ssl
            self.use_keepalive = config.use_keepalive

        self.conn = None

        self.token = token

        if shared_data is None:
            self.shared_data = SimpleNamespace()
            self.shared_data.session = None
            self.shared_data.lock = Lock()
        else:
            self.shared_data = shared_data

        self.should_authenticate = authenticate
        # One time flag to indicate if we ever connected,
        # to prevent logging of connection errors on first connect.
        self._ever_connected = False

    def authenticate(self, shared_data, user, password, token):
        """
        Authenticate with the database. This will be called automatically from query.
        This is separate from session refresh mechanism, and is called only once.
        Failure leads to exception.
        """
        if not self.authenticated:
            if shared_data.session is None:
                self.shared_data.lock = Lock()
                try:
                    self._authenticate(user, password, token)
                except Exception as e:
                    raise UnauthenticatedException(
                        "Authentication failed:", str(e))
            else:
                self.shared_data = shared_data
            self.authenticated = True

    def __del__(self):
        if self.connected:
            self.conn.close()
            self.connected = False

    def _send_msg(self, data):
        sent_len = struct.pack('@I', len(data))  # send size first
        x = self.conn.send(sent_len + data)
        return x == len(data) + 4

    def _recv_msg(self):
        recv_len = self.conn.recv(4)  # get message size
        if recv_len == b'':
            return None
        recv_len = struct.unpack('@I', recv_len)[0]
        response = bytearray(recv_len)
        read = 0
        while read < recv_len:
            packet = self.conn.recv(recv_len - read)
            if recv_len == b'':
                return None
            if not packet:
                logger.error("Error receiving")
                return None
            response[read:] = packet
            read += len(packet)

        return response

    def _authenticate(self, user, password="", token=""):

        query = [{
            "Authenticate": {
                "username": user,
            }
        }]

        if password:
            query[0]["Authenticate"]["password"] = password
        elif token:
            query[0]["Authenticate"]["token"] = token
        else:
            raise Exception(
                "Either password or token must be specified for authentication")

        response, _ = self._query(query)

        if not isinstance(response, (list, tuple)
                          ) or "Authenticate" not in response[0]:
            raise Exception(
                "Unexpected response from server upon authenticate request: " +
                str(response))
        session_info = response[0]["Authenticate"]
        if session_info["status"] != 0:
            raise Exception(session_info["info"])

        self.shared_data.session = Session(
            session_info["session_token"],
            session_info["refresh_token"],
            session_info["session_token_expires_in"],
            session_info["refresh_token_expires_in"],
            time.time())

    def _check_session_status(self):
        if not self.shared_data.session:
            return

        if not self.shared_data.session.valid():
            with self.shared_data.lock:
                self._refresh_token()

    def _refresh_token(self):
        query = [{
            "RefreshToken": {
                "refresh_token": self.shared_data.session.refresh_token
            }
        }]

        response, _ = self._query(query, [], try_resume=False)

        logger.info(f"Refresh token response: \r\n{response}")
        if isinstance(response, list):
            session_info = response[0]["RefreshToken"]
            if session_info["status"] != 0:
                raise UnauthorizedException(response)

            self.shared_data.session = Session(
                session_info["session_token"],
                session_info["refresh_token"],
                session_info["session_token_expires_in"],
                session_info["refresh_token_expires_in"],
                time.time())
        else:
            raise UnauthorizedException(response)

    def _connect(self):

        self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.conn.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        if self.use_keepalive:
            keepalive.set(self.conn)

        # TCP_QUICKACK only supported in Linux 2.4.4+.
        # We use startswith for checking the platform following Python's
        # documentation:
        # https://docs.python.org/dev/library/sys.html#sys.platform
        if sys.platform.startswith('linux'):
            self.conn.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)

        try:
            self.conn.connect((self.host, self.port))

            # Handshake with server to negotiate protocol

            protocol = 2 if self.use_ssl else 1

            hello_msg = struct.pack('@II', PROTOCOL_VERSION, protocol)

            # Send desire protocol
            self._send_msg(hello_msg)

            # Receive response from server
            response = self._recv_msg()

            version, server_protocol = struct.unpack('@II', response)

            if version != PROTOCOL_VERSION:
                logger.warning("Protocol version differ from server")

            if server_protocol != protocol:
                self.conn.close()
                self.connected = False
                raise Exception(
                    "Server did not accept protocol. Aborting Connection.")

            if self.use_ssl:

                # Server is ok with SSL, we switch over SSL.
                self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                self.context.check_hostname = False
                # TODO, we need to add support for local certificates
                # For now, we let the server send us the certificate
                self.context.verify_mode = ssl.VerifyMode.CERT_NONE
                self.conn = self.context.wrap_socket(self.conn)

        except BaseException as e:
            logger.error(
                f"Error connecting to server: {str(e)} {self.config}",
                exc_info=True,
                stack_info=True)
            self.conn.close()
            self.connected = False
            raise

        self.connected = True

    def connect(self, details: str = None):
        self.connected = False
        if not self.connected:
            # Connection is not established, keep trying.
            # nothing else will work until this is done.
            # This can help to get going amidst server restarts.
            try:
                self._connect()
            except socket.error as e:
                logger.error(
                    f"Error connecting to server: {self.config} \r\n{details}.",
                    exc_info=True,
                    stack_info=True)

    def _query(self, query, blob_array = [], try_resume=True):
        response_blob_array = []
        # Check the query type
        if not isinstance(query, str):  # assumes json
            query_str = json.dumps(query)
        else:
            query_str = query

        query_msg = queryMessage_pb2.queryMessage()
        # query has .json and .blobs
        query_msg.json = query_str

        # Set Auth token, only when not authenticated before
        if self.shared_data.session:
            query_msg.token = self.shared_data.session.session_token

        for blob in blob_array:
            query_msg.blobs.append(blob)

        # Serialize with protobuf and send
        data = query_msg.SerializeToString()

        # this is for session refresh attempts
        tries = 0
        while tries < self.config.retry_max_attempts:
            try:
                if self._send_msg(data):
                    response = self._recv_msg()
                    if response is not None:
                        querRes = queryMessage_pb2.queryMessage()
                        querRes.ParseFromString(response)
                        response_blob_array = [b for b in querRes.blobs]
                        self.last_response = json.loads(querRes.json)
                        break
            except ssl.SSLEOFError as ssle:
                # this can happen when working in a notebook.
                # we log if this isn't the first try, or if
                # it has happened sooner than we expect a connection to be
                # dropped.
                now = datetime.now()
                if tries != 0 or (self.last_query_timestamp is not None and
                                  (now - self.last_query_timestamp) <
                                  self.query_connection_error_suppression_delta):
                    logger.exception(ssle)
                    logger.warning(
                        f"SSL connection error on process {os.getpid()}")
            except ssl.SSLError as ssle:
                # This can happen in a scenario where multiple
                # processes might be accessing a single connection.
                # The copy does not make usable connections.
                logger.exception(ssle)
                logger.warning(f"SSL error on process {os.getpid()}")
            except OSError as ose:
                logger.exception(ose)
                logger.warning(f"OS error on process {os.getpid()}")
            except AttributeError as ae:
                if self.connected:
                    # Only log if we got this while connected.
                    # else it is expected after unification of query/connect
                    logger.exception(ae)
                    logger.warning(f"Attribute error on process {os.getpid()}")

            tries += 1
            # Do not log when trying for the first time.
            if self._ever_connected:
                logger.warning(
                    f"Connection broken. Reconnecting attempt [{tries}/{self.config.retry_max_attempts}] .. PID = {os.getpid()}")

            if self.connected:
                self.conn.close()
                self.connected = False

            self.connect(
                details=f"Will retry in {self.config.retry_interval_seconds} seconds")
            if not self._ever_connected:
                self._ever_connected = True
            time.sleep(self.config.retry_interval_seconds)

            # Try to resume the session, in cases where the connection is severed.
            # For example aperturedb server is restarted, or network is lost.
            # While this is useful bit of code, when executed in a refresh token
            # path, this can cause a deadlock. Hence the try_resume flag.
            if try_resume:
                self._renew_session()
        if tries == self.config.retry_max_attempts:
            raise Exception(
                f"Could not query apertureDB using TCP.")
        return (self.last_response, response_blob_array)

    def query(self, q, blobs=[]):
        """
        Query the database with a query string or a json object.
        First it checks if the session is valid, if not, it refreshes the token.
        Then it sends the query to the server and returns the response.

        Args:
            q (json): native query to be sent
            blobs (list, optional): Blobs if needed with the query. Defaults to [].

        Raises:
            ConnectionError: Fatal error, connection to server lost

        Returns:
            _type_: _description_
        """
        if self.should_authenticate:
            self.authenticate(
                shared_data=self.shared_data,
                user=self.config.username,
                password=self.config.password,
                token=self.token)

        try:
            start = time.time()
            self.response, self.blobs = self._query(q, blobs)
            if not isinstance(
                    self.response,
                    list) and self.response["info"] == "Not Authenticated!":
                # The case where session is valid, but expires while query is sent.
                # Hope is that the query send won't be longer than the session
                # ttl.
                logger.warning(
                    f"Session expired while query was sent. Retrying... {self.config}")
                self._renew_session()
                start = time.time()
                self.response, self.blobs = self._query(q, blobs)
            self.last_query_time = time.time() - start
            self.last_query_timestamp = datetime.now()
            return self.response, self.blobs
        except BaseException as e:
            logger.critical("Failed to query",
                            exc_info=True, stack_info=True)
            raise

    def _renew_session(self):
        count = 0
        while count < 3:
            try:
                self._check_session_status()
                break
            except UnauthorizedException as e:
                logger.warning(
                    f"[Attempt {count + 1} of 3] Failed to refresh token.",
                    exc_info=True,
                    stack_info=True)
                time.sleep(1)
                count += 1

    def create_new_connection(self) -> Connector:
        """
        Create a new connection object with the same parameters as the current one.
        This is important in multi-threaded applications, where each thread should have its own connection object.
        Because the connection object is not thread-safe, it is not safe to share it between threads.
        Be cautious when using this method, as it will create a new connection to the database, which will consume resources.
        Ideally, this method should be called once for each thread.

        Returns:
            connection: Clone of original connection
        """
        return type(self)(
            self.host,
            self.port,
            self.config.username,
            self.config.password,
            self.token,
            use_ssl=self.use_ssl,
            shared_data=self.shared_data)

    def get_last_response_str(self):

        return json.dumps(self.last_response, indent=4, sort_keys=False)

    def print_last_response(self):

        print(self.get_last_response_str())

    def get_last_query_time(self):

        return self.last_query_time

    def get_response(self):

        return self.response

    def get_blobs(self):

        return self.blobs

    # Return a flag on whether the last query succeeded:
    # i.e., all commands return status >= 0
    def last_query_ok(self):

        return self.check_status(self.response) >= 0

    def check_status(self, json_res):

        status = 0
        if (isinstance(self.response, dict)):
            if ("status" not in self.response):
                status = self.check_status(json_res[list(json_res.keys())[0]])
            else:
                status = json_res["status"]
        elif (isinstance(json_res, (tuple, list))):
            if ("status" not in json_res[0]):
                status = self.check_status(json_res[0])
            else:
                status = json_res[0]["status"]

        return status



from aperturedb import CSVParser
import logging

logger = logging.getLogger(__name__)
# SparseAddingDataCSV
# Check for item existance using constraints before adding
# Useful when adding larger resources where a portion already exist


class SparseAddingDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB General CSV Parser for Loading Blob data where a large amount of the blobs already exist.

    This is a blob loader where the entity is searched for first, before the blob data is passed to the server.
    This can be useful speedup if blob data is large in comparison to the amount of data actually causing loads

    This is an abstract class, ImageSparseAddDataCSV loads Images.

    """

    def __init__(self, entity_class: str, filename: str, **kwargs):
        self.entity = entity_class
        self.keys_set = False
        super().__init__(filename, **kwargs)
        self.blobs_per_query = [0, 1]
        self.commands_per_query = [1, 1]
        self._setupkeys()

    def _setupkeys(self):
        if not self.keys_set:
            self.keys_set = True
            self.props_keys = [x for x in self.header[1:]
                               if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
            self.constraints_keys = [x for x in self.header[1:]
                                     if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]

    def getitem(self, idx):
        idx = self.df.index.start + idx
        query_set = []

        hold_props_keys = self.props_keys
        self.props_keys = []
        self.command = "Find" + self.entity
        self.constraint_keyword = "constraints"
        entity_find = self._basic_command(
            idx, custom_fields={"results": {"count": True}})
        # proceed to second command if count == 0
        condition_find_failed = {"results": {0: {"count": ["==", 0]}}}
        self.props_keys = hold_props_keys
        self.command = "Add" + self.entity
        self.constraint_keyword = "if_not_found"
        entity_add = self._basic_command(idx)
        query_set.append(entity_find)
        query_set.append([condition_find_failed, entity_add])

        if hasattr(self, "modify_item") and callable(self.modify_item):
            query_set = self.modify_item(query_set, idx)

        return [query_set], []

    def validate(self):
        self._setupkeys()
        valid = True
        if not self.use_dask:
            if len(self.constraints_keys) < 1:
                logger.error("Cannot add/update " +
                             self.entity + "; no constraint keys")
                valid = False
        return valid



from __future__ import annotations


class Operations(object):
    """
    **Operations that can be performed on the fly on any retrieved images**

    [Supported operations](/query_language/Reference/shared_command_parameters/operations)
    """

    def __init__(self):

        self.operations_arr = []

    def get_operations_arr(self):
        return self.operations_arr

    def resize(self, width: int, height: int) -> Operations:

        op = {
            "type": "resize",
            "width":  width,
            "height": height,
        }

        self.operations_arr.append(op)
        return self

    def rotate(self, angle: int, resize=False) -> Operations:

        op = {
            "type": "rotate",
            "angle": angle,
            "resize": resize,
        }

        self.operations_arr.append(op)
        return self

    def flip(self, code: str) -> Operations:

        op = {
            "type": "flip",
            "code": code,
        }

        self.operations_arr.append(op)
        return self

    def crop(self, x: int, y: int, width: int, height: int) -> Operations:

        op = {
            "type": "crop",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
        }

        self.operations_arr.append(op)
        return self

    def interval(self, start: int, stop: int, step: int) -> Operations:

        op = {
            "type": "interval",
            "start": start,
            "stop": stop,
            "step": step
        }

        self.operations_arr.append(op)
        return self
